# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import re
from dataclasses import dataclass
from itertools import product

import kernel_explorer as ke
import numpy as np
import pytest
from utils import dtype_to_bytes, root_mean_square, standardization


def get_bert_sizes_test():
    batch_sizes = [1, 8]
    seq_lens = [64, 256]
    hidden_sizes = [1, 2, 3, 4, 5, 7, 8, 9, 13, 32, 63, 64, 65, 127, 128, 129, 177, 256, 1023, 1024]
    return product(batch_sizes, seq_lens, hidden_sizes)


def get_bert_sizes_profile():
    batch_sizes = [1, 8, 128, 256]
    seq_lens = [64, 128, 256, 384]
    hidden_sizes = [768, 1024]
    return product(batch_sizes, seq_lens, hidden_sizes)


def dtype_to_funcs(dtype, simplified=False):
    skip_layer_norm_prefix = "SimplifiedSkipLayerNorm" if simplified else "SkipLayerNorm"
    type_map = {
        "float16": list(filter(lambda x: re.match(f"{skip_layer_norm_prefix}.*_half.*", x), dir(ke))),
        "float32": list(filter(lambda x: re.match(f"{skip_layer_norm_prefix}.*_float.*", x), dir(ke))),
    }
    return type_map[dtype]


def skip_layer_norm(input_x, skip, bias, gamma, beta, epsilon):
    val = input_x + skip + bias
    output = standardization(val, 2, epsilon)
    output = output * gamma + beta
    return output, val


def simplified_skip_layer_norm(input_x, skip, bias, gamma, epsilon):
    val = input_x + skip + bias
    rms = root_mean_square(val, 2, epsilon)
    output = (val / rms) * gamma
    return output, val


@ke.dispatchable(pattern_arg=4)
def run_skip_layer_norm(
    batch_size: int, seq_len: int, hidden_size: int, dtype: str, func, simplified=False, has_optional_output=False
):
    np.random.seed(0)
    input_x = np.random.rand(batch_size, seq_len, hidden_size).astype(dtype)
    skip = np.random.rand(batch_size, seq_len, hidden_size).astype(dtype)
    bias = np.random.rand(hidden_size).astype(dtype)
    gamma = np.random.rand(hidden_size).astype(dtype)
    beta = np.random.rand(hidden_size).astype(dtype)
    # Because of rocm FMAs calculation issue with float16, epsilon should be larger when hidden_size is small
    epsilon = 0.05 if hidden_size < 8 else 0.0005
    output_y = np.random.rand(batch_size, seq_len, hidden_size).astype(dtype)
    output_optional = (
        np.random.rand(batch_size, seq_len, hidden_size).astype(dtype)
        if has_optional_output
        else np.empty((0), dtype=dtype)
    )

    input_d = ke.DeviceArray(input_x)
    skip_d = ke.DeviceArray(skip)
    bias_d = ke.DeviceArray(bias)
    gamma_d = ke.DeviceArray(gamma)
    beta_d = ke.DeviceArray(beta)
    y_d = ke.DeviceArray(output_y)
    optional_d = ke.DeviceArray(output_optional)
    f = getattr(ke, func)

    my_op = f(
        y_d,
        optional_d,
        input_d,
        skip_d,
        gamma_d,
        beta_d,
        bias_d,
        epsilon,
        hidden_size,
        batch_size * seq_len * hidden_size,
    )
    if my_op.IsSupported():
        my_op.Run()

        y_d.UpdateHostNumpyArray()
        optional_d.UpdateHostNumpyArray()

        if simplified:
            y_ref, y_optional = simplified_skip_layer_norm(input_x, skip, bias, gamma, epsilon)
        else:
            y_ref, y_optional = skip_layer_norm(input_x, skip, bias, gamma, beta, epsilon)
        np.testing.assert_almost_equal(y_ref, output_y, decimal=1)
        if has_optional_output:
            np.testing.assert_almost_equal(y_optional, output_optional, decimal=3)


dtypes = ["float32", "float16"]
simplified = [True, False]


@pytest.mark.parametrize("bert_sizes", get_bert_sizes_test())
@pytest.mark.parametrize("dtype", dtypes)
@pytest.mark.parametrize("simplified", simplified)
def test_skip_layer_norm(bert_sizes, dtype, simplified):
    for func in dtype_to_funcs(dtype, simplified):
        run_skip_layer_norm(*bert_sizes, dtype, func, simplified)


@dataclass
class SkipLayerNormMetric(ke.BandwidthMetric):
    batch_size: int
    seq_len: int
    hidden_size: int

    def report(self):
        common = f"{self.dtype}  batch_size={self.batch_size:<4} seq_len={self.seq_len:<4} hidden_size={self.hidden_size:<4} {self.name}"
        if self.duration > 0:
            return f"{self.duration:6.2f} us, {self.gbps:5.2f} GB/s " + common
        return "not supported          " + common


@ke.dispatchable(pattern_arg=4)
def profile_skip_layer_norm_func(batch_size, seq_len, hidden_size, dtype, func, has_optional_output):
    np.random.seed(0)
    input_x = np.random.rand(batch_size, seq_len, hidden_size).astype(dtype)
    skip = np.random.rand(batch_size, seq_len, hidden_size).astype(dtype)
    gamma = np.random.rand(hidden_size).astype(dtype)
    beta = np.random.rand(hidden_size).astype(dtype)
    bias = np.random.rand(hidden_size).astype(dtype)
    epsilon = 0.0005
    output_y = np.random.rand(batch_size, seq_len, hidden_size).astype(dtype)
    output_optional = (
        np.random.rand(batch_size, seq_len, hidden_size).astype(dtype)
        if has_optional_output
        else np.empty((0), dtype=dtype)
    )

    input_d = ke.DeviceArray(input_x)
    skip_d = ke.DeviceArray(skip)
    gamma_d = ke.DeviceArray(gamma)
    beta_d = ke.DeviceArray(beta)
    bias_d = ke.DeviceArray(bias)
    y_d = ke.DeviceArray(output_y)
    optional_d = ke.DeviceArray(output_optional)
    f = getattr(ke, func)

    my_op = f(
        y_d,
        optional_d,
        input_d,
        skip_d,
        gamma_d,
        beta_d,
        bias_d,
        epsilon,
        hidden_size,
        batch_size * seq_len * hidden_size,
    )

    duration_ms = -1
    if my_op.IsSupported():
        duration_ms = my_op.Profile()
    total_bytes = (input_x.size * 3 + bias.size * 3) * dtype_to_bytes(dtype)

    ke.report(SkipLayerNormMetric(func, dtype, duration_ms, total_bytes, batch_size, seq_len, hidden_size))


@ke.dispatchable
def profile_with_args(batch_size, seq_len, hidden_size, dtype, has_optional_output=False, simplified=False):
    with ke.benchmark():
        for func in dtype_to_funcs(dtype, simplified):
            profile_skip_layer_norm_func(batch_size, seq_len, hidden_size, dtype, func, has_optional_output)


def profile():
    for dtype in dtypes:
        for bert_size in get_bert_sizes_profile():
            profile_with_args(*bert_size, dtype)
            print()


if __name__ == "__main__":
    parser = ke.get_argument_parser()
    group = parser.add_argument_group()
    group.add_argument("batch_size", type=int)
    group.add_argument("seq_len", type=int)
    group.add_argument("hidden_size", type=int)
    group.add_argument("dtype", choices=dtypes)
    group.add_argument("--has_optional_output", "-o", action="store_true")
    group.add_argument("--simplified", "-s", action="store_true", default=False)

    if not ke.has_args():
        profile()
    else:
        args = parser.parse_args()
        args.dispatch(
            args.batch_size,
            args.seq_len,
            args.hidden_size,
            args.dtype,
            args.has_optional_output,
            args.simplified,
        )



# https://docs.microsoft.com/en-us/azure/mysql/connect-python

import argparse
import os
from datetime import datetime

import git
import mysql.connector
from mysql.connector import errorcode


def get_repo_commit(repo_path):
    repo = git.Repo(repo_path, search_parent_directories=True)
    sha = repo.head.object.hexsha
    short_sha = repo.git.rev_parse(sha, short=4)
    return short_sha


create_table_script = "CREATE TABLE perf_test_training_ort_module_data (\
    id int(11) NOT NULL AUTO_INCREMENT,\
    Model varchar(64) COLLATE utf8_bin DEFAULT NULL,\
    BatchId varchar(32) COLLATE utf8_bin DEFAULT NULL,\
    CommitId varchar(32) COLLATE utf8_bin DEFAULT NULL,\
    ModelName varchar(256) COLLATE utf8_bin DEFAULT NULL,\
    DisplayName varchar(512) COLLATE utf8_bin DEFAULT NULL,\
    UseMixedPrecision tinyint(1) DEFAULT NULL,\
    UseAutoCast tinyint(1) DEFAULT NULL,\
    UseDeepSpeed tinyint(1) DEFAULT NULL,\
    Optimizer varchar(32) COLLATE utf8_bin DEFAULT NULL,\
    BatchSize int(11) DEFAULT NULL,\
    SeqLen int(11) DEFAULT NULL,\
    PredictionsPerSeq int(11) DEFAULT NULL,\
    NumOfBatches int(11) DEFAULT NULL,\
    WeightUpdateSteps int(11) DEFAULT NULL,\
    Round int(11) DEFAULT NULL,\
    GradAccSteps int(11) DEFAULT NULL,\
    AvgTimePerBatch float DEFAULT NULL,\
    Throughput float DEFAULT NULL,\
    StabilizedThroughput float DEFAULT NULL,\
    EndToEndThroughput float DEFAULT NULL,\
    TotalTime float DEFAULT NULL,\
    AvgCPU int(11) DEFAULT NULL,\
    Memory int(11) DEFAULT NULL,\
    RunConfig varchar(2048) COLLATE utf8_bin DEFAULT NULL,\
    Time datetime DEFAULT NULL,\
    PRIMARY KEY (id),\
    UNIQUE KEY config_unique (Model,BatchId,CommitId,UseMixedPrecision,UseAutoCast,UseDeepSpeed,Optimizer,BatchSize,SeqLen,ModelName)\
) ENGINE=InnoDB AUTO_INCREMENT=1696 DEFAULT CHARSET=utf8 COLLATE=utf8_bin;"

insert_table_script = "INSERT INTO onnxruntime.perf_test_training_ort_module_data\
    (\
    Model,\
    BatchId,\
    CommitId,\
    ModelName,\
    DisplayName,\
    UseMixedPrecision,\
    UseAutoCast,\
    UseDeepSpeed,\
    Optimizer,\
    BatchSize,\
    SeqLen,\
    PredictionsPerSeq,\
    NumOfBatches,\
    WeightUpdateSteps,\
    Round,\
    GradAccSteps,\
    AvgTimePerBatch,\
    Throughput,\
    StabilizedThroughput,\
    EndToEndThroughput,\
    TotalTime,\
    AvgCPU,\
    Memory,\
    RunConfig,\
    Time)\
    VALUES\
    (\
    %(Model)s,\
    %(BatchId)s,\
    %(CommitId)s,\
    %(ModelName)s,\
    %(DisplayName)s,\
    %(UseMixedPrecision)s,\
    %(UseAutoCast)s,\
    %(UseDeepSpeed)s,\
    %(Optimizer)s,\
    %(BatchSize)s,\
    %(SeqLen)s,\
    %(PredictionsPerSeq)s,\
    %(NumOfBatches)s,\
    %(WeightUpdateSteps)s,\
    %(Round)s,\
    %(GradAccSteps)s,\
    %(AvgTimePerBatch)s,\
    %(Throughput)s,\
    %(StabilizedThroughput)s,\
    %(EndToEndThroughput)s,\
    %(TotalTime)s,\
    %(AvgCPU)s,\
    %(Memory)s,\
    %(RunConfig)s,\
    %(Time)s)"


# Obtain connection string information from the portal
def connect_to_perf_dashboard_db(mysql_server_name, power_bi_user_name, password, database):
    config = {
        "host": mysql_server_name,
        "user": power_bi_user_name,
        "password": password,
        "database": database,
    }

    try:
        conn = mysql.connector.connect(**config)
        print("Connection established")
        return conn
    except mysql.connector.Error as err:
        if err.errno == errorcode.ER_ACCESS_DENIED_ERROR:
            print("Something is wrong with the user name or password")
        elif err.errno == errorcode.ER_BAD_DB_ERROR:
            print("Database does not exist")
        else:
            print(err)


def log_perf_metrics(
    perf_metrics, mysql_server_name, power_bi_user_name, power_bi_password, power_bi_database, perf_repo_path=None
):
    if perf_repo_path:
        perf_metrics["CommitId"] = get_repo_commit(perf_repo_path)
    else:
        perf_metrics["CommitId"] = get_repo_commit(os.path.realpath(__file__))

    connect_and_insert_perf_metrics(
        mysql_server_name, power_bi_user_name, power_bi_password, power_bi_database, perf_metrics
    )


required_attributes_for_perf_metrics = [
    "model_name",
    "optimizer",
    "batch_size",
    "epochs",
    "train_steps",
    "sequence_length",
]


def calculate_and_log_perf_metrics(
    args, start_time, mysql_server_name, power_bi_user_name, power_bi_password, power_bi_database, ort_repo_path=None
):
    completion_time = datetime.datetime.now()
    perf_metrics_duration = completion_time - start_time

    for attribute in required_attributes_for_perf_metrics:
        if not hasattr(args, attribute):
            raise ValueError(
                "args does not contain all attributes needed to calculate perf metrics. \
                Please prepare perf_metrics and call log_perf_metrics instead"
            )

    perf_metrics = {}
    perf_metrics["Model"] = args.model_name
    perf_metrics["BatchId"] = "NA"
    perf_metrics["ModelName"] = args.model_name
    perf_metrics["DisplayName"] = args.model_name
    perf_metrics["UseMixedPrecision"] = args.fp16 if hasattr(args, "fp16") else False
    perf_metrics["UseAutoCast"] = args.use_auto_cast if hasattr(args, "use_auto_cast") else False
    perf_metrics["UseDeepSpeed"] = args.use_deep_speed if hasattr(args, "use_deep_speed") else False
    perf_metrics["Optimizer"] = args.optimizer
    perf_metrics["BatchSize"] = args.batch_size
    perf_metrics["SeqLen"] = args.sequence_length
    perf_metrics["PredictionsPerSeq"] = args.prediction_per_seq if hasattr(args, "prediction_per_seq") else 0
    perf_metrics["NumOfBatches"] = args.epochs * args.train_steps
    perf_metrics["WeightUpdateSteps"] = args.epochs * args.train_steps
    perf_metrics["Round"] = 0  # NA
    perf_metrics["GradAccSteps"] = args.gradient_accumulation_steps

    perf_metrics["AvgTimePerBatch"] = perf_metrics_duration.microseconds / args.train_steps

    perf_metrics["Throughput"] = args.batch_size * args.train_steps / perf_metrics_duration.seconds

    perf_metrics["StabilizedThroughput"] = 0  # TODO
    perf_metrics["EndToEndThroughput"] = 0  # TODO
    perf_metrics["TotalTime"] = perf_metrics_duration.seconds

    perf_metrics["AvgCPU"] = 0  # TODO
    perf_metrics["Memory"] = 0  # TODO
    perf_metrics["RunConfig"] = "na"
    perf_metrics["Time"] = completion_time.strftime("%Y-%m-%d %H:%M:%S")

    log_perf_metrics(
        perf_metrics, mysql_server_name, power_bi_user_name, power_bi_password, power_bi_database, ort_repo_path
    )


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("--mysql_server_name", help="Perf dashboard mysql server name")
    parser.add_argument("--power_bi_user_name", help="Power BI user name")
    parser.add_argument("--password", help="password", default=None)
    parser.add_argument("--database", help="The dashboard database")
    return parser.parse_args()


def connect_and_insert_perf_metrics(mysql_server_name, power_bi_user_name, password, database, perf_metrics):
    conn = connect_to_perf_dashboard_db(mysql_server_name, power_bi_user_name, password, database)
    # https://dev.mysql.com/doc/connector-python/en/connector-python-api-mysqlcursor-execute.html
    conn.cursor().execute(insert_table_script, perf_metrics)
    conn.commit()
    conn.cursor().close()
    conn.close()
    print("perf_metrics logged into power-bi database.")


if __name__ == "__main__":
    args = parse_arguments()
    conn = connect_to_perf_dashboard_db(args.mysql_server_name, args.power_bi_user_name, args.password, args.database)
    conn.cursor().execute(create_table_script)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------


from ._modifier_registry import OptimizerModifierTypeRegistry, get_full_qualified_type_name


def FP16_Optimizer(optimizer, **kwargs):  # noqa: N802
    """
    Simple wrapper to replace inefficient FP16_Optimizer function calls implemented by libraries for example
        Apex, DeepSpeed, Megatron-LM.

    Usage:
        1. DeepSpeed ZeRO Optimizer Override:

        >>> from onnxruntime.training.optim.fp16_optimizer import FP16_Optimizer
        >>> optimizer = Adam(param_groups,
        >>>                  lr=args.lr,
        >>>                  weight_decay=args.weight_decay,
        >>>                  betas=(args.adam_beta1, args.adam_beta2),
        >>>                  eps=args.adam_eps)

        >>> model, optimizer, _, lr_scheduler = deepspeed.initialize(
        >>>     model=model,
        >>>     optimizer=optimizer,
        >>>     args=args,
        >>>     lr_scheduler=lr_scheduler,
        >>>     mpu=mpu,
        >>>     dist_init_required=False)
        >>> if args.fp16:
        >>>     optimizer = FP16_Optimizer(optimizer)

        2. Megatron-LM-v1.1.5 Optimizer Override:

        >>> from onnxruntime.training.optim.fp16_optimizer import FP16_Optimizer as ORT_FP16_Optimizer
        >>> optimizer = Adam(param_groups,
        >>>                  lr=args.lr,
        >>>                  weight_decay=args.weight_decay,
        >>>                  betas=(args.adam_beta1, args.adam_beta2),
        >>>                  eps=args.adam_eps)

        >>> # Wrap into fp16 optimizer.
        >>> if args.fp16:
        >>>     optimizer = FP16_Optimizer(optimizer,
        >>>                                static_loss_scale=args.loss_scale,
        >>>                                dynamic_loss_scale=args.dynamic_loss_scale,
        >>>                                dynamic_loss_args={
        >>>                                     'scale_window': args.loss_scale_window,
        >>>                                     'min_scale': args.min_scale,
        >>>                                     'delayed_shift': args.hysteresis},
        >>>                                verbose=True)
        >>>     optimizer = ORT_FP16_Optimizer(optimizer,
        >>>                                    get_tensor_model_parallel_rank=mpu.get_model_parallel_rank,
        >>>                                    get_tensor_model_parallel_group=mpu.get_model_parallel_group)

        3. APEX AMP Override:

        >>> from onnxruntime.training.optim.fp16_optimizer import FP16_Optimizer as ORT_FP16_Optimizer
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=1e-3)

        >>> model, optimizer = amp.initialize(model, optimizer, opt_level="O2")
        >>> optimizer = ORT_FP16_Optimizer(optimizer)
        >>>
        >>> # Wrap model with ORTModule tricks
        >>> def patch_new_fwd(old_new_fwd):
        >>>     def new_new_fwd(self, *args, **kwargs):
        >>>         return old_new_fwd(*args, **kwargs)
        >>>     return new_new_fwd

        >>> model.forward = types.MethodType(patch_new_fwd(model.forward), model)
        >>> model = ORTModule(model)
    Args:
        optimizer: the FP16_Optimizer instance

    Returns:
        The modified FP16_Optimizer instance

    """

    optimizer_full_qualified_name = (
        "apex.amp.optimizer.unique_name_as_id"
        if hasattr(optimizer, "_amp_stash")
        else get_full_qualified_type_name(optimizer)
    )
    modifier = OptimizerModifierTypeRegistry.create_modifier(optimizer_full_qualified_name, optimizer, **kwargs)
    if modifier is not None:
        modifier.apply()

    return optimizer



from pathlib import Path

import numpy as np
import onnx

HERE = Path(__file__).parent.resolve(strict=True)
TEST = False

if TEST:
    import onnxruntime


def generate_fuse_pad_avgpool():
    parameters = {
        "fuse-pad-avgpool": (
            {},
            [[1.333333, 2.333333, 1.777778], [3.0, 5.0, 3.666667], [2.666667, 4.333333, 3.111111]],
        ),
        "fuse-pad-avgpool_with_pad": (
            {"pads": [1, 1, 0, 0], "count_include_pad": 1},
            [
                [0.111111, 0.333333, 0.666667, 0.555556],
                [0.555556, 1.333333, 2.333333, 1.777778],
                [1.333333, 3.0, 5.0, 3.666667],
                [1.222222, 2.666667, 4.333333, 3.111111],
            ],
        ),
        "fuse-pad-avgpool_with_pad-nofuse": (
            {"pads": [1, 1, 0, 0]},
            [
                [0.25, 0.5, 1.0, 0.833333],
                [0.833333, 1.333333, 2.333333, 1.777778],
                [2.0, 3.0, 5.0, 3.666667],
                [1.833333, 2.666667, 4.333333, 3.111111],
            ],
        ),
    }
    for name in parameters:
        model_path = HERE / f"{name}.onnx"
        input_ = onnx.helper.make_tensor_value_info("input", onnx.TensorProto.FLOAT, (1, 1, 3, 3))
        pad = onnx.helper.make_node("Pad", ["input"], ["tp"], mode="constant", pads=[0, 0, 1, 1, 0, 0, 1, 1])
        pool = onnx.helper.make_node("AveragePool", ["tp"], ["output"], kernel_shape=[3, 3], **parameters[name][0])
        nodes = [pad, pool]
        output_shape = (1, 1, 3, 3) if name == "fuse-pad-avgpool" else (1, 1, 4, 4)
        output_ = onnx.helper.make_tensor_value_info("output", onnx.TensorProto.FLOAT, output_shape)
        graph = onnx.helper.make_graph(nodes, name, [input_], [output_])
        model = onnx.helper.make_model(graph, opset_imports=[onnx.helper.make_opsetid("", 7)])
        onnx.checker.check_model(model)
        onnx.save_model(model, model_path)
        if TEST:
            input_array = np.array([[[[1, 2, 3], [4, 5, 6], [7, 8, 9]]]], dtype=np.float32)
            expected = np.array(parameters[name][1], dtype=np.float32)
            session_options = onnxruntime.SessionOptions()
            session_options.execution_mode = onnxruntime.ExecutionMode.ORT_SEQUENTIAL
            session_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_DISABLE_ALL
            session = onnxruntime.InferenceSession(model_path, session_options)
            out = session.run(["output"], {"input": input_array})
            actual = out[0].squeeze()
            np.testing.assert_allclose(actual, expected, rtol=1e-5, atol=0.0)
            session_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
            session = onnxruntime.InferenceSession(model_path, session_options)
            out = session.run(["output"], {"input": input_array})
            actual = out[0].squeeze()
            np.testing.assert_allclose(actual, expected, rtol=1e-5, atol=0.0)


if __name__ == "__main__":
    generate_fuse_pad_avgpool()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging

import torch

logger = logging.getLogger(__name__)


class WhisperDecoderInitOpenai(torch.nn.Module):
    """WhisperDecoderInit for Openai."""

    def __init__(
        self,
        model: torch.nn.Module,
        decoder: torch.nn.Module,
    ):
        super().__init__()
        self.whisper_model = model
        self.whisper_decoder = decoder
        self.kv_cache = {}

    @torch.no_grad()
    def forward(
        self,
        tokens,
        audio_features,
        past=None,
        remove_hooks=False,
    ):
        # Create a kv_cache for past_values
        past_kv_cache = {}
        if past is not None:
            # Convert past values from 4D to 3D
            past = [torch.transpose(val, 1, 2) for val in past]
            past = [val.reshape(val.shape[:2] + (-1,)) for val in past]
            half_idx = len(past) // 2
            for idx, block in enumerate(self.whisper_decoder.blocks):
                past_kv_cache[block.attn.key] = past[2 * idx]
                past_kv_cache[block.attn.value] = past[2 * idx + 1]
                past_kv_cache[block.cross_attn.key] = past[2 * idx + half_idx]
                past_kv_cache[block.cross_attn.value] = past[2 * idx + half_idx + 1]

        hooks = None
        if not self.kv_cache:
            self.kv_cache, hooks = self.whisper_model.install_kv_cache_hooks()

        logits = self.whisper_decoder(tokens, audio_features, kv_cache=past_kv_cache)

        # Add concat node for past values
        if past is not None:
            for block in self.whisper_decoder.blocks:
                self.kv_cache[block.attn.key] = torch.cat(
                    [past_kv_cache[block.attn.key], self.kv_cache[block.attn.key]], dim=1
                ).detach()
                self.kv_cache[block.attn.value] = torch.cat(
                    [past_kv_cache[block.attn.value], self.kv_cache[block.attn.value]], dim=1
                ).detach()

        present_self, present_cross = [], []
        # Group self and cross values
        for block in self.whisper_decoder.blocks:
            present_self.append(self.kv_cache[block.attn.key])
            present_self.append(self.kv_cache[block.attn.value])
            if past is None:
                present_cross.append(self.kv_cache[block.cross_attn.key])
                present_cross.append(self.kv_cache[block.cross_attn.value])

        present_self = present_self + present_cross
        # Add reshape and transpose ops to convert from 3D to 4D
        present_self = [
            present_val.reshape(present_val.shape[:2] + (-1, 64)).transpose(1, 2) for present_val in present_self
        ]

        # Remove forward hooks to avoid model cloning step
        if hooks is not None and remove_hooks:
            self.kv_cache = {}
            for hook in hooks:
                hook.remove()
        return logits, present_self



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import numpy as np

DebugOutput = False
np.set_printoptions(suppress=True)  # , precision=16, floatmode='maxprec')


def print_with_shape(name, a, force_output=False):
    if force_output or DebugOutput:
        print(name + " [shape: ", a.shape, "]\n", a)


def print_results(Y):
    print("*************************")
    print_with_shape("Y", Y, True)
    print("*************************")


class GRU_Helper:  # noqa: N801
    def __init__(self, **params):
        # Match the ONNXRuntime/CNTK behavior
        # If False use the python from the ONNX spec
        self.match_onnxruntime = True

        required_inputs = ["X", "W", "R"]
        for i in required_inputs:
            assert i in params, f"Missing Required Input: {i}"

        num_directions = params["W"].shape[0]
        params["X"].shape[0]

        hidden_size = params["R"].shape[-1]
        batch_size = params["X"].shape[1]

        X = params["X"]  # noqa: N806
        W = params["W"]  # noqa: N806
        R = params["R"]  # noqa: N806
        B = (  # noqa: N806
            params["B"]
            if "B" in params
            else np.zeros(num_directions * 6 * hidden_size).reshape(num_directions, 6 * hidden_size)
        )
        H_0 = (  # noqa: N806
            params["initial_h"]
            if "initial_h" in params
            else np.zeros((num_directions, batch_size, hidden_size)).reshape(num_directions, batch_size, hidden_size)
        )
        LBR = params.get("linear_before_reset", 0)  # noqa: N806
        self.direction = params.get("direction", "forward")

        if num_directions == 1:
            if self.direction == "forward":
                self.one = OneDirectionGRU(X, W, R, B, H_0, LBR)
            else:
                # flip input so we process in reverse
                self.one = OneDirectionGRU(np.flip(X, 0), W, R, B, H_0, LBR)

            self.two = None

        else:
            # split the inputs which have per direction rows
            Wfw, Wbw = np.vsplit(W, 2)  # noqa: N806
            Rfw, Rbw = np.vsplit(R, 2)  # noqa: N806
            Bfw, Bbw = np.vsplit(B, 2)  # noqa: N806
            H_0fw, H_0bw = np.vsplit(H_0, 2)  # noqa: N806

            self.one = OneDirectionGRU(X, Wfw, Rfw, Bfw, H_0fw, LBR)
            self.two = OneDirectionGRU(np.flip(X, 0), Wbw, Rbw, Bbw, H_0bw, LBR)

    def run(self):
        if self.direction == "bidirectional":
            f_output = self.one.execute()
            r_output = self.two.execute()

            # flip reverse output it matches the original input order
            r_output_orig_input_order = np.flip(r_output, 0)

            # create merged output by merging the forward and reverse rows for seq_length
            # 0 rows, 2 directions, batch size, hidden_size
            seq_length = f_output.shape[0]
            batch_size = f_output.shape[2]
            hidden_size = f_output.shape[3]

            output = np.empty((0, 2, batch_size, hidden_size), np.float32)
            for x in range(seq_length):
                output = np.append(output, f_output[x])
                output = np.append(output, r_output_orig_input_order[x])

            output = output.reshape(seq_length, 2, batch_size, hidden_size)
        else:
            output = self.one.execute()
            if self.direction == "reverse":
                # flip so it's back in the original order of the inputs
                output = np.flip(output, 0)

        return output


class OneDirectionGRU:
    def __init__(self, X, W, R, B, initial_h, LBR):
        self.X = X
        # remove num_directions axis for W, R, B, H_0
        self.W = np.squeeze(W, axis=0)
        self.R = np.squeeze(R, axis=0)
        self.B = np.squeeze(B, axis=0)
        self.H_0 = np.squeeze(initial_h, axis=0)
        self.LBR = LBR

    def f(self, x):
        return 1 / (1 + np.exp(-x))

    def g(self, x):
        return np.tanh(x)

    def execute(self):
        print_with_shape("X", self.X)

        [w_z, w_r, w_h] = np.split(self.W, 3)
        [r_z, r_r, r_h] = np.split(self.R, 3)
        [w_bz, w_br, w_bh, r_bz, r_br, r_bh] = np.split(self.B, 6)

        # print_with_shape("w_z", w_z)
        # print_with_shape("w_r", w_r)
        # print_with_shape("w_h", w_h)

        # print_with_shape("r_z", r_z)
        # print_with_shape("r_r", r_r)
        # print_with_shape("r_h", r_h)

        # print_with_shape("w_bz", w_bz)
        # print_with_shape("w_br", w_br)
        # print_with_shape("w_bh", w_bh)
        # print_with_shape("r_bz", r_bz)
        # print_with_shape("r_br", r_br)
        # print_with_shape("r_bh", r_bh)

        self.X.shape[0]
        num_directions = 1
        hidden_size = self.R.shape[-1]
        batch_size = self.X.shape[1]

        output = np.empty((0, num_directions, batch_size, hidden_size), np.float32)

        for row in self.X:
            z = self.f(np.dot(row, np.transpose(w_z)) + np.dot(self.H_0, np.transpose(r_z)) + w_bz + r_bz)
            r = self.f(np.dot(row, np.transpose(w_r)) + np.dot(self.H_0, np.transpose(r_r)) + w_br + r_br)
            h_default = self.g(np.dot(row, np.transpose(w_h)) + np.dot(r * self.H_0, np.transpose(r_h)) + w_bh + r_bh)
            h_linear = self.g(np.dot(row, np.transpose(w_h)) + r * (np.dot(self.H_0, np.transpose(r_h)) + r_bh) + w_bh)

            h = h_linear if self.LBR else h_default

            print_with_shape("z", z)
            print_with_shape("r", r)
            print_with_shape("h", h)

            H = (1 - z) * h + z * self.H_0  # noqa: N806

            print_with_shape("H", H)
            output = np.append(output, H.reshape(1, 1, batch_size, hidden_size), axis=0)

            self.H_0 = H

        return output


class ONNXRuntimeTestContext:
    @staticmethod
    def OneDirectionWeights():  # noqa: N802
        hidden_size = 2

        W = np.array(  # noqa: N806
            [
                [
                    [-0.494659, 0.0453352],  # Wz
                    [-0.487793, 0.417264],
                    [-0.0091708, -0.255364],  # Wr
                    [-0.106952, -0.266717],
                    [-0.0888852, -0.428709],  # Wh
                    [-0.283349, 0.208792],
                ]
            ]
        ).astype(np.float32)

        R = np.array(  # noqa: N806
            [
                [
                    [0.146626, -0.0620289],  # Rz
                    [-0.0815302, 0.100482],
                    [-0.228172, 0.405972],  # Rr
                    [0.31576, 0.281487],
                    [-0.394864, 0.42111],  # Rh
                    [-0.386624, -0.390225],
                ]
            ]
        ).astype(np.float32)

        W_B = np.array(  # noqa: N806
            [
                [
                    0.381619,
                    0.0323954,
                    -0.258721,
                    0.45056,
                    -0.250755,
                    0.0967895,
                ]
            ]
        ).astype(  # Wbz  # Wbr
            np.float32
        )  # Wbh
        R_B = np.zeros((1, 3 * hidden_size)).astype(np.float32)  # noqa: N806
        B = np.concatenate((W_B, R_B), axis=1)  # noqa: N806

        return W, R, B

    @staticmethod
    def BidirectionalWeights():  # noqa: N802
        W1, R1, B1 = ONNXRuntimeTestContext.OneDirectionWeights()  # noqa: N806

        hidden_size = R1.shape[-1]
        input_size = W1.shape[-1]

        W = np.tile(W1, (2, 1)).reshape(2, 3 * hidden_size, input_size)  # noqa: N806
        R = np.tile(R1, (2, 1)).reshape(2, 3 * hidden_size, hidden_size)  # noqa: N806
        B = np.tile(B1, (2, 1))  # noqa: N806

        return W, R, B


# replicate ONNXRuntime unit tests inputs to validate output
class GRU_ONNXRuntimeUnitTests:  # noqa: N801
    @staticmethod
    def ForwardDefaultActivationsSimpleWeightsNoBiasTwoRows():  # noqa: N802
        print(GRU_ONNXRuntimeUnitTests.ForwardDefaultActivationsSimpleWeightsNoBiasTwoRows.__name__)

        seq_length = 2
        batch_size = 2
        input_size = 1
        hidden_size = 3
        input = np.array([1.0, 2.0, 10.0, 11.0]).astype(np.float32).reshape(seq_length, batch_size, input_size)

        W = (  # noqa: N806
            np.array([0.1, 0.2, 0.3, 1, 2, 3, 10, 11, 12]).astype(np.float32).reshape(1, 3 * hidden_size, input_size)
        )

        weight_scale = 0.1
        R = weight_scale * np.ones((1, 3 * hidden_size, hidden_size)).astype(np.float32)  # noqa: N806

        gru = GRU_Helper(X=input, W=W, R=R, direction="forward")
        fw_output = gru.run()
        print_results(fw_output)

    @staticmethod
    def ReverseDefaultActivationsSimpleWeightsNoBiasTwoRows():  # noqa: N802
        print(GRU_ONNXRuntimeUnitTests.ReverseDefaultActivationsSimpleWeightsNoBiasTwoRows.__name__)

        input_size = 1
        hidden_size = 3
        input = np.array([[[1.0], [2.0]], [[10.0], [11.0]]]).astype(np.float32)

        W = (  # noqa: N806
            np.array([0.1, 0.2, 0.3, 1, 2, 3, 10, 11, 12]).astype(np.float32).reshape(1, 3 * hidden_size, input_size)
        )

        weight_scale = 0.1
        R = weight_scale * np.ones((1, 3 * hidden_size, hidden_size)).astype(np.float32)  # noqa: N806

        gru = GRU_Helper(X=input, W=W, R=R, direction="reverse")
        fw_output = gru.run()
        print_results(fw_output)

    @staticmethod
    def BidirectionalDefaultActivationsSimpleWeightsNoBias(linear_before_reset=0):  # noqa: N802
        print(
            GRU_ONNXRuntimeUnitTests.BidirectionalDefaultActivationsSimpleWeightsNoBias.__name__
            + ".linear_before_reset="
            + str(linear_before_reset)
        )

        input_size = 1
        hidden_size = 3

        if linear_before_reset:
            input = np.array([[[1.0], [2.0], [3.0]], [[10.0], [11.0], [12.0]]]).astype(np.float32)
        else:
            input = np.array([[[1.0], [2.0]], [[10.0], [11.0]]]).astype(np.float32)

        W = (  # noqa: N806
            np.array([0.1, 0.2, 0.3, 1, 2, 3, 10, 11, 12]).astype(np.float32).reshape(1, 3 * hidden_size, input_size)
        )

        weight_scale = 0.1
        R = weight_scale * np.ones((1, 3 * hidden_size, hidden_size)).astype(np.float32)  # noqa: N806

        # duplicate the W and R inputs so we use the same values for both forward and reverse
        gru = GRU_Helper(
            X=input,
            W=np.tile(W, (2, 1)).reshape(2, 3 * hidden_size, input_size),
            R=np.tile(R, (2, 1)).reshape(2, 3 * hidden_size, hidden_size),
            direction="bidirectional",
            linear_before_reset=linear_before_reset,
        )

        fw_output = gru.run()
        print_results(fw_output)

    @staticmethod
    def DefaultActivationsSimpleWeightsWithBias(rows=2, direction="forward", linear_before_reset=0):  # noqa: N802
        print(
            GRU_ONNXRuntimeUnitTests.DefaultActivationsSimpleWeightsWithBias.__name__
            + " batch_parallel="
            + str(rows != 1)
            + " direction="
            + direction
            + " linear_before_reset="
            + str(linear_before_reset)
        )

        seq_length = 2
        batch_size = rows
        input_size = 1
        hidden_size = 3

        if batch_size == 1:
            input = [-0.1, -0.3]
        else:
            input = [-0.1, 0.2, -0.3, 0.4]

        input = np.array(input).astype(np.float32).reshape(seq_length, batch_size, input_size)

        W = (  # noqa: N806
            np.array([0.1, 0.2, 0.3, 0.2, 0.3, 0.1, 0.3, 0.1, 0.2])
            .astype(np.float32)
            .reshape(1, 3 * hidden_size, input_size)
        )

        weight_scale = 0.1
        R = weight_scale * np.ones((1, 3 * hidden_size, hidden_size)).astype(np.float32)  # noqa: N806

        # Wb[zrh] Rb[zrh]
        B = (  # noqa: N806
            np.array(
                [
                    -0.01,
                    0.1,
                    0.01,
                    -0.2,
                    -0.02,
                    0.02,
                    0.3,
                    -0.3,
                    -0.3,
                    -0.03,
                    0.5,
                    -0.7,
                    0.05,
                    -0.7,
                    0.3,
                    0.07,
                    -0.03,
                    0.5,
                ]
            )
            .astype(np.float32)
            .reshape(1, 6 * hidden_size)
        )

        gru = GRU_Helper(
            X=input,
            W=W,
            R=R,
            B=B,
            direction=direction,
            linear_before_reset=linear_before_reset,
        )
        fw_output = gru.run()
        print_results(fw_output)

    @staticmethod
    def ForwardDefaultActivationsSimpleWeightsWithBiasBatchParallel():  # noqa: N802
        GRU_ONNXRuntimeUnitTests.DefaultActivationsSimpleWeightsWithBias()

    @staticmethod
    def ForwardDefaultActivationsSimpleWeightsWithBiasBatchParallelLinearBeforeReset():  # noqa: N802
        GRU_ONNXRuntimeUnitTests.DefaultActivationsSimpleWeightsWithBias(linear_before_reset=1)

    @staticmethod
    def ReverseDefaultActivationsSimpleWeightsWithBiasBatchParallelLinearBeforeReset():  # noqa: N802
        GRU_ONNXRuntimeUnitTests.DefaultActivationsSimpleWeightsWithBias(direction="reverse", linear_before_reset=1)

    @staticmethod
    def ForwardDefaultActivationsSimpleWeightsWithBiasLinearBeforeReset():  # noqa: N802
        GRU_ONNXRuntimeUnitTests.DefaultActivationsSimpleWeightsWithBias(rows=1, linear_before_reset=1)

    @staticmethod
    def ReverseDefaultActivationsSimpleWeightsWithBiasLinearBeforeReset():  # noqa: N802
        GRU_ONNXRuntimeUnitTests.DefaultActivationsSimpleWeightsWithBias(
            rows=1, direction="reverse", linear_before_reset=1
        )

    @staticmethod
    def Legacy_TestGRUOpForwardBasic():  # noqa: N802
        print(GRU_ONNXRuntimeUnitTests.Legacy_TestGRUOpForwardBasic.__name__)

        input = np.array([[[-0.455351, -0.276391]], [[-0.185934, -0.269585]]]).astype(np.float32)

        W, R, B = ONNXRuntimeTestContext.OneDirectionWeights()  # noqa: N806
        gru = GRU_Helper(X=input, W=W, R=R, B=B)
        output = gru.run()
        print_results(output)

    @staticmethod
    def Legacy_TestGRUOpBackwardBasic():  # noqa: N802
        print(GRU_ONNXRuntimeUnitTests.Legacy_TestGRUOpBackwardBasic.__name__)

        input = np.array([[[-0.185934, -0.269585]], [[-0.455351, -0.276391]]]).astype(np.float32)

        W, R, B = ONNXRuntimeTestContext.OneDirectionWeights()  # noqa: N806
        gru = GRU_Helper(X=input, W=W, R=R, B=B, direction="reverse")
        output = gru.run()
        print_results(output)

    @staticmethod
    def Legacy_TestGRUOpBidirectionalBasic():  # noqa: N802
        print(GRU_ONNXRuntimeUnitTests.Legacy_TestGRUOpBidirectionalBasic.__name__)

        input = np.array([[[-0.455351, -0.276391]], [[-0.185934, -0.269585]]]).astype(np.float32)

        W, R, B = ONNXRuntimeTestContext.BidirectionalWeights()  # noqa: N806
        gru = GRU_Helper(X=input, W=W, R=R, B=B, direction="bidirectional")
        output = gru.run()
        print_results(output)


GRU_ONNXRuntimeUnitTests.ForwardDefaultActivationsSimpleWeightsNoBiasTwoRows()
GRU_ONNXRuntimeUnitTests.ReverseDefaultActivationsSimpleWeightsNoBiasTwoRows()
GRU_ONNXRuntimeUnitTests.BidirectionalDefaultActivationsSimpleWeightsNoBias()
GRU_ONNXRuntimeUnitTests.BidirectionalDefaultActivationsSimpleWeightsNoBias(linear_before_reset=1)

GRU_ONNXRuntimeUnitTests.ForwardDefaultActivationsSimpleWeightsWithBiasBatchParallel()
GRU_ONNXRuntimeUnitTests.ForwardDefaultActivationsSimpleWeightsWithBiasBatchParallelLinearBeforeReset()
GRU_ONNXRuntimeUnitTests.ReverseDefaultActivationsSimpleWeightsWithBiasBatchParallelLinearBeforeReset()
GRU_ONNXRuntimeUnitTests.ForwardDefaultActivationsSimpleWeightsWithBiasLinearBeforeReset()
GRU_ONNXRuntimeUnitTests.ReverseDefaultActivationsSimpleWeightsWithBiasLinearBeforeReset()

GRU_ONNXRuntimeUnitTests.Legacy_TestGRUOpForwardBasic()
GRU_ONNXRuntimeUnitTests.Legacy_TestGRUOpBackwardBasic()
GRU_ONNXRuntimeUnitTests.Legacy_TestGRUOpBidirectionalBasic()



import argparse

import deepspeed
import torch
from deepspeed.pipe import LayerSpec
from torch import nn, utils

from onnxruntime.training.ortmodule.experimental.pipe import ORTPipelineModule

# This script demonstrates how to set up a pipeline parallel training session
# using DeepSpeed's ORTPipelineModule for a simple neural network model.


# USAGE:
# pip install deepspeed
# deepspeed orttraining_test_ort_pipeline_module.py --deepspeed_config=orttraining_test_ortmodule_deepspeed_pipeline_parallel_config.json --pipeline-parallel-size 2 --steps=100
def get_args() -> argparse.Namespace:
    """Parse and return command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--local_rank", type=int, default=-1, help="Local rank passed from distributed launcher")
    parser.add_argument("--steps", type=int, default=100, help="Number of training steps to run")
    parser.add_argument("--pipeline-parallel-size", type=int, default=2, help="Number of pipeline stages")
    parser.add_argument("--backend", type=str, default="nccl", help="Distributed backend")
    parser.add_argument("--seed", type=int, default=0, help="Random seed for reproducibility")
    parser.add_argument("--layer_spec", type=bool, default=False, help="Use LayerSpec for layer specification")

    parser = deepspeed.add_config_arguments(parser)
    return parser.parse_args()


class SampleData(utils.data.Dataset):
    """Custom dataset to facilitate loading and batching of data."""

    def __init__(self, x: torch.Tensor, y: torch.Tensor):
        self.x = x
        self.y = y

    def __len__(self) -> int:
        return self.x.size(0)

    def __getitem__(self, idx: int) -> tuple[torch.Tensor, torch.Tensor]:
        return self.x[idx], self.y[idx]


class SimpleNetPipeInput(nn.Module):
    """First stage of the pipeline, responsible for initial processing."""

    def __init__(self, config: dict[str, int]):
        super().__init__()
        self.linear = nn.Linear(config["input_size"], config["hidden_size"])
        self.activation = nn.ReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear(x)
        x = self.activation(x)
        return x


class SimpleNetPipeBlock(nn.Module):
    """Intermediate stage of the pipeline, can be duplicated to deepen the network."""

    def __init__(self, config: dict[str, int]):
        super().__init__()
        self.linear = nn.Linear(config["hidden_size"], config["hidden_size"])
        self.activation = nn.ReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear(x)
        x = self.activation(x)
        return x


class SimpleNetPipeOutput(nn.Module):
    """Final stage of the pipeline, producing the output."""

    def __init__(self, config: dict[str, int]):
        super().__init__()
        self.linear = nn.Linear(config["hidden_size"], config["output_size"])

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear(x)
        return x


def build_model(config: dict[str, int], n: int, layer_spec: bool) -> nn.Module:
    """Constructs and returns the model either using LayerSpec or nn.Sequential."""
    if layer_spec:
        print("Wrapping layers with LayerSpec")
        model = (
            [LayerSpec(SimpleNetPipeInput, config)]
            + [LayerSpec(SimpleNetPipeBlock, config) for _ in range(n)]
            + [LayerSpec(SimpleNetPipeOutput, config)]
        )
    else:
        print("Wrapping layers with nn.Sequential")
        model = nn.Sequential(
            SimpleNetPipeInput(config),
            SimpleNetPipeBlock(config),
            SimpleNetPipeBlock(config),
            SimpleNetPipeOutput(config),
        )
    return model


args = get_args()
torch.cuda.set_device(args.local_rank)
device = torch.device("cuda", args.local_rank)
deepspeed.init_distributed(dist_backend=args.backend)
torch.manual_seed(args.seed)

model = build_model({"input_size": 4, "hidden_size": 8, "output_size": 3}, n=10, layer_spec=args.layer_spec)

model = ORTPipelineModule(
    layers=model,
    loss_fn=torch.nn.CrossEntropyLoss(),
    num_stages=args.pipeline_parallel_size,
    partition_method="uniform",
    activation_checkpoint_interval=0,
)

# Setup input data
x = torch.rand((10, 4))
y = torch.randint(0, 3, (10,))
dataset = SampleData(x, y)

print("Initialize deepspeed")
model_engine, optimizer, _, _ = deepspeed.initialize(
    args=args, model=model, model_parameters=model.parameters(), training_data=dataset
)

for step in range(args.steps):
    loss = model_engine.train_batch()
    if step % 10 == 0:
        print(f"step = {step}, loss = {loss}")



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import os
import unittest

import coloredlogs
import pytest
from parity_utilities import find_transformers_source

if find_transformers_source(sub_dir_paths=["models", "gpt2"]):
    from benchmark_gpt2 import main, parse_arguments
else:
    from onnxruntime.transformers.models.gpt2.benchmark_gpt2 import main, parse_arguments


class TestGpt2(unittest.TestCase):
    def setUp(self):
        from onnxruntime import get_available_providers

        self.test_cuda = "CUDAExecutionProvider" in get_available_providers()

    def run_benchmark_gpt2(self, arguments: str):
        args = parse_arguments(arguments.split())
        csv_filename = main(args)
        self.assertIsNotNone(csv_filename)
        self.assertTrue(os.path.exists(csv_filename))

    @pytest.mark.slow
    def test_gpt2_stage1(self):
        self.run_benchmark_gpt2("-m gpt2 --precision fp32 -v -b 1 --sequence_lengths 8 -s 0 --stage 1")

    @pytest.mark.slow
    def test_gpt2_fp32(self):
        self.run_benchmark_gpt2("-m gpt2 --precision fp32 -v -b 1 --sequence_lengths 2 -s 3")

    @pytest.mark.slow
    def test_gpt2_fp16(self):
        if self.test_cuda:
            self.run_benchmark_gpt2("-m gpt2 --precision fp16 -o -b 1 --sequence_lengths 2 -s 3 --use_gpu")

    @pytest.mark.slow
    def test_gpt2_int8(self):
        self.run_benchmark_gpt2("-m gpt2 --precision int8 -o  -b 1 --sequence_lengths 2 -s 3")


if __name__ == "__main__":
    coloredlogs.install(fmt="%(message)s")
    logging.getLogger("transformers").setLevel(logging.ERROR)
    unittest.main()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

class AttributeType(object):
    UNDEFINED = 0
    FLOAT = 1
    INT = 2
    STRING = 3
    TENSOR = 4
    GRAPH = 5
    FLOATS = 6
    INTS = 7
    STRINGS = 8
    TENSORS = 9
    GRAPHS = 10
    SPARSE_TENSOR = 11
    SPARSE_TENSORS = 12



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

from dataclasses import dataclass
from itertools import product

import kernel_explorer as ke
import numpy as np
import pytest
from utils import (
    dtype_to_suffix,
    fast_gelu,
    get_gemm_basic_sizes,
    get_gemm_bert_sizes,
    get_gemm_bound,
    matmul,
    transab_to_suffix,
)


# TODO The test method needs update.
def _test_gemmfastgelu(my_func, dtype: str, m: int, n: int, k: int, transa=False, transb=False):
    assert dtype in ["float16", "float32"]

    a_shape = (k, m) if transa else (m, k)
    b_shape = (n, k) if transb else (k, n)

    np.random.seed(0)
    a = (np.random.rand(*a_shape)).astype(dtype).astype("float64")
    b = (np.random.rand(*b_shape)).astype(dtype).astype("float64")
    bias = (np.random.rand(n)).astype(dtype)
    temp_c = matmul(a, b, transa, transb)

    bound = get_gemm_bound(dtype, a, b, temp_c, transa, transb, a_b_positive=True)

    temp_c = temp_c.astype(dtype)
    ref_c = fast_gelu(temp_c, bias)

    a = a.astype(dtype)
    b = b.astype(dtype)

    my_c = np.zeros((m, n), dtype=dtype)
    dev_a = ke.DeviceArray(a)
    dev_b = ke.DeviceArray(b)
    dev_bias = ke.DeviceArray(bias)
    dev_c = ke.DeviceArray(my_c)

    opa = ke.blas_op.T if transa else ke.blas_op.N
    opb = ke.blas_op.T if transb else ke.blas_op.N
    lda = a_shape[1]
    ldb = b_shape[1]
    alpha = 1.0
    beta = 0.0
    my_op = my_func(opa, opb, m, n, k, alpha, dev_a, lda, dev_b, ldb, dev_bias, beta, dev_c, n)

    print(f"dtype={dtype} {transab_to_suffix((transa, transb))} m={m:<5} n={n:<5} k={k:<5} bound: {max(bound, 1e-2)}")

    for impl in my_op.ListOps():
        if not my_op.SelectOp(impl):
            continue

        my_op.Run()
        dev_c.UpdateHostNumpyArray()

        np.testing.assert_allclose(my_c, ref_c, rtol=max(bound, 1e-2))


dtypes = ["float16", "float32"]
all_transabs = list(product([True, False], repeat=2))


@pytest.mark.parametrize("dtype", dtypes)
@pytest.mark.parametrize("size", get_gemm_basic_sizes(full=False) + get_gemm_bert_sizes(full=False))
@pytest.mark.parametrize("transab", all_transabs)
def test_gemmfastgelu_unfused_bert_cases(dtype, size, transab):
    _test_gemmfastgelu(getattr(ke, "GemmFastGeluUnfused_" + dtype_to_suffix(dtype)), dtype, *size, *transab)


@pytest.mark.parametrize("dtype", dtypes)
@pytest.mark.parametrize("size", get_gemm_basic_sizes(full=False) + get_gemm_bert_sizes(full=False))
@pytest.mark.parametrize("transab", all_transabs)
def test_gemmfastgelu_tunable_bert_cases(dtype, size, transab):
    wrapper_name = f"GemmFastGeluTunable_{dtype_to_suffix(dtype)}_{transab_to_suffix(transab)}"
    _test_gemmfastgelu(getattr(ke, wrapper_name), dtype, *size, *transab)


@pytest.mark.skipif(not ke.is_composable_kernel_available(), reason="ck is not enabled")
@pytest.mark.parametrize("dtype", dtypes)
@pytest.mark.parametrize("size", get_gemm_basic_sizes(full=False) + get_gemm_bert_sizes(full=False))
@pytest.mark.parametrize("transab", all_transabs)
def test_gemmfastgelu_ck_bert_cases(dtype, size, transab):
    wrapper_name = f"CKGemmFastGelu_{dtype_to_suffix(dtype)}_{transab_to_suffix(transab)}"
    _test_gemmfastgelu(getattr(ke, wrapper_name), dtype, *size, *transab)


@pytest.mark.skipif(not ke.is_hipblaslt_available(), reason="hipblaslt is not available")
@pytest.mark.parametrize("dtype", dtypes)
@pytest.mark.parametrize("size", get_gemm_basic_sizes(full=False) + get_gemm_bert_sizes(full=False))
@pytest.mark.parametrize("transab", all_transabs)
def test_gemmfastgelu_hipblaslt_bert_cases(dtype, size, transab):
    _test_gemmfastgelu(getattr(ke, "GemmFastGeluHipBlasLt_" + dtype_to_suffix(dtype)), dtype, *size, *transab)


@dataclass
class GemmFastGeluMetric(ke.ComputeMetric):
    transa: bool
    transb: bool
    m: int
    n: int
    k: int

    def report(self):
        transab = transab_to_suffix((self.transa, self.transb))
        common = f"{self.dtype} m={self.m:<4} n={self.n:<4} k={self.k:<4} {transab}, {self.name}"
        if self.duration <= 0:
            return "not supported          " + common
        return f"{self.duration:>6.2f} us {self.tflops:>5.2f} tflops " + common


@ke.dispatchable(pattern_arg=0)
def profile_gemmfastgelu_func(my_func, dtype: str, m: int, n: int, k: int, transa: bool, transb: bool):
    a_shape = (k, m) if transa else (m, k)
    b_shape = (n, k) if transb else (k, n)

    np.random.seed(0)
    a = (np.random.rand(*a_shape) * 2 - 1).astype(dtype)
    b = (np.random.rand(*b_shape) * 2 - 1).astype(dtype)
    my_c = np.zeros((m, n), dtype=dtype)
    bias = np.random.rand(n).astype(dtype)

    dev_a = ke.DeviceArray(a)
    dev_b = ke.DeviceArray(b)
    dev_bias = ke.DeviceArray(bias)
    dev_c = ke.DeviceArray(my_c)

    opa = ke.blas_op.T if transa else ke.blas_op.N
    opb = ke.blas_op.T if transb else ke.blas_op.N
    lda = a_shape[1]
    ldb = b_shape[1]
    alpha = 1.0
    beta = 0.0
    my_op = my_func(opa, opb, m, n, k, alpha, dev_a, lda, dev_b, ldb, dev_bias, beta, dev_c, n)

    for impl in my_op.ListOps():
        duration_ms = -1
        if my_op.SelectOp(impl):
            duration_ms = my_op.Profile()
        # only counts gemm tflops because fastgelu is low order term (7 * n).
        floating_point_operations = m * k * n * 2

        ke.report(GemmFastGeluMetric(impl, dtype, duration_ms, floating_point_operations, transa, transb, m, n, k))


@ke.dispatchable
def profile_with_args(transa, transb, dtype, m, n, k):
    dtype_suffix = "_" + dtype_to_suffix(dtype)
    transab_suffix = "_" + transab_to_suffix((transa, transb))
    with ke.benchmark():
        profile_gemmfastgelu_func(getattr(ke, "GemmFastGeluUnfused" + dtype_suffix), dtype, m, n, k, transa, transb)
        profile_gemmfastgelu_func(
            getattr(ke, "CKGemmFastGelu" + dtype_suffix + transab_suffix), dtype, m, n, k, transa, transb
        )
        profile_gemmfastgelu_func(
            getattr(ke, "GemmFastGeluTunable" + dtype_suffix + transab_suffix), dtype, m, n, k, transa, transb
        )
        if ke.is_hipblaslt_available():
            profile_gemmfastgelu_func(
                getattr(ke, "GemmFastGeluHipBlasLt" + dtype_suffix + transab_suffix), dtype, m, n, k, transa, transb
            )


def profile():
    for dtype in dtypes:
        for m, n, k in get_gemm_bert_sizes(full=True):
            profile_with_args(False, False, dtype, m, n, k)
            print()


if __name__ == "__main__":
    parser = ke.get_argument_parser()
    group = parser.add_argument_group()
    group.add_argument("transa", choices="NT")
    group.add_argument("transb", choices="NT")
    group.add_argument("dtype", choices=dtypes)
    group.add_argument("m", type=int)
    group.add_argument("n", type=int)
    group.add_argument("k", type=int)

    if not ke.has_args():
        profile()
    else:
        args = parser.parse_args()
        args.dispatch(args.transa == "T", args.transb == "T", args.dtype, args.m, args.n, args.k)



import sys


def read_values(filename):
    with open(filename) as file:
        values = [(float(line.strip()), i + 1) for i, line in enumerate(file)]
    return values


def top_n(values, N):
    return sorted(values, key=lambda x: x[0], reverse=True)[:N]


def compare_files(cpu_file, npu_file, N):
    cpu_values = read_values(cpu_file)
    npu_values = read_values(npu_file)

    cpu_topn = top_n(cpu_values, N)
    npu_topn = top_n(npu_values, N)

    print(f"Top-{N} values in {cpu_file}: {cpu_topn}")
    print(f"Top-{N} values in {npu_file}: {npu_topn}")


if __name__ == "__main__":
    if len(sys.argv) != 4:
        print("Usage: python compare_topn.py <N> <cpu_file> <npu_file>")
        sys.exit(1)

    N = int(sys.argv[1])
    cpu_file = sys.argv[2]
    npu_file = sys.argv[3]

    compare_files(cpu_file, npu_file, N)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import contextlib
import inspect
import os
import sys
import warnings

import torch
from packaging import version

from onnxruntime import set_seed
from onnxruntime.capi import build_and_package_info as ort_info
from onnxruntime.capi._pybind_state import is_ortmodule_available

from ._fallback import ORTModuleFallbackException, ORTModuleInitException, _FallbackPolicy, wrap_exception
from .torch_cpp_extensions import is_installed as is_torch_cpp_extensions_installed

if not is_ortmodule_available():
    raise ImportError("ORTModule is not supported on this platform.")


def _defined_from_envvar(name: str, default_value: any, warn: bool = True):
    """Check given name exists in the environment variable and return the value using the default_value's
    type if it exists.
    """
    new_value = os.getenv(name, None)
    if new_value is None:
        return default_value
    try:
        new_value = type(default_value)(new_value)
    except (TypeError, ValueError) as e:
        if warn:
            warnings.warn(f"Unable to overwrite constant {name!r} due to {e!r}.")
        return default_value
    return new_value


def _override_gradient_checkpoint(original_checkpoint):
    """
    Best effort to override `torch.utils.checkpoint` and `deepspeed.checkpointing.checkpoint` during ONNX export.

    Despite importing `torch.utils.checkpoint` or `deepspeed.checkpointing.checkpoint` in `__init__.py`,
    users might import it first, causing our override to not take effect. We still attempt to override
    it to work in most cases.

    We replace the checkpoint function with our implementation, without condition checks.
    The actual check is in the overridden function, verifying if:
    1) `checkpoint` is called during ORTModule model export,
    2) Gradient checkpoint autograd function is disallowed (ORTMODULE_ALLOW_AUTOGRAD_CHECKPOINT),
    3) Memory optimization level is not specified by the user (ORTMODULE_MEMORY_OPT_LEVEL).
    If true, we reset memory optimization to layer-wise recompute.

    """

    # Note: The `torch.utils.checkpoint` checkpoint function signature looks like below:
    #   `checkpoint(function, *args,
    #               use_reentrant = None,
    #               context_fn = noop_context_fn,
    #               determinism_check = _DEFAULT_DETERMINISM_MODE,
    #               debug = False,
    #               **kwargs).`
    # The few keyword arguments are not used in the recompute module forward function, but by the
    # checkpoint function itself, so we need to filter them out otherwise module forward function
    # would complain about unexpected keyword arguments.
    all_input_parameters = inspect.signature(original_checkpoint).parameters.values()
    outside_kwarg_params = []
    for input_parameter in all_input_parameters:
        if (
            input_parameter.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
            or input_parameter.kind == inspect.Parameter.KEYWORD_ONLY
            or input_parameter.kind == inspect.Parameter.VAR_KEYWORD
        ):
            outside_kwarg_params.append(input_parameter.name)

    def _checkpoint(
        function,
        *args,
        **kwargs,
    ):
        # Conditions to activate layer-wise memory optimization automatically:
        # 1. Checkpoint is called during ORTModule model export context.
        # 2. Gradient checkpoint autograd function export is disallowed.
        # 3. Memory optimization level is layer-wise recompute.
        if (
            ORTMODULE_ONNX_EXPORT_CONTEXT[0] is True
            and _defined_from_envvar("ORTMODULE_ALLOW_AUTOGRAD_CHECKPOINT", 0) != 1
            and _defined_from_envvar("ORTMODULE_MEMORY_OPT_LEVEL", 0) == 1
        ):
            for name in outside_kwarg_params:
                if name in kwargs:
                    # Pop out the keyword argument to avoid passing it to the module run function
                    kwargs.pop(name)
            print(
                "Layer-wise memory optimization is enabled upon detecting "
                "gradient checkpointing autograd function usage during model execution."
            )
            return function(*args, **kwargs)
        return original_checkpoint(
            function,
            *args,
            **kwargs,
        )

    return _checkpoint


with contextlib.suppress(Exception):
    from torch.utils.checkpoint import checkpoint as original_torch_checkpoint

    torch.utils.checkpoint.checkpoint = _override_gradient_checkpoint(original_torch_checkpoint)

    import deepspeed

    original_deepspeed_checkpoint = deepspeed.checkpointing.checkpoint
    deepspeed.checkpointing.checkpoint = _override_gradient_checkpoint(original_deepspeed_checkpoint)


################################################################################
# All global constant goes here, before ORTModule is imported ##################
# NOTE: To *change* values in runtime, import onnxruntime.training.ortmodule and
# assign them new values. Importing them directly do not propagate changes.
################################################################################
ONNX_OPSET_VERSION = 17
MINIMUM_RUNTIME_PYTORCH_VERSION_STR = "1.8.1"
ORTMODULE_TORCH_CPP_DIR = os.path.join(os.path.dirname(__file__), "torch_cpp_extensions")
_FALLBACK_INIT_EXCEPTION = None
ORTMODULE_FALLBACK_POLICY = (
    _FallbackPolicy.FALLBACK_UNSUPPORTED_DEVICE
    | _FallbackPolicy.FALLBACK_UNSUPPORTED_DATA
    | _FallbackPolicy.FALLBACK_UNSUPPORTED_TORCH_MODEL
    | _FallbackPolicy.FALLBACK_UNSUPPORTED_ONNX_MODEL
)
ORTMODULE_FALLBACK_RETRY = False
ORTMODULE_IS_DETERMINISTIC = torch.are_deterministic_algorithms_enabled()

ONNXRUNTIME_CUDA_VERSION = ort_info.cuda_version if hasattr(ort_info, "cuda_version") else None
ONNXRUNTIME_ROCM_VERSION = ort_info.rocm_version if hasattr(ort_info, "rocm_version") else None

# The first value indicates whether the code is in ONNX export context.
# The export context here include the full export process, including prepare export input/output information,
# and export model.
ORTMODULE_ONNX_EXPORT_CONTEXT = [False]


@contextlib.contextmanager
def export_context():
    """Context manager for model export."""
    try:
        ORTMODULE_ONNX_EXPORT_CONTEXT[0] = True

        yield
    finally:
        ORTMODULE_ONNX_EXPORT_CONTEXT[0] = False


# Verify minimum PyTorch version is installed before proceeding to ONNX Runtime initialization
try:
    import torch

    runtime_pytorch_version = version.parse(torch.__version__.split("+")[0])
    minimum_runtime_pytorch_version = version.parse(MINIMUM_RUNTIME_PYTORCH_VERSION_STR)
    if runtime_pytorch_version < minimum_runtime_pytorch_version:
        raise wrap_exception(
            ORTModuleInitException,
            RuntimeError(
                "ONNX Runtime ORTModule frontend requires PyTorch version greater"
                f" or equal to {MINIMUM_RUNTIME_PYTORCH_VERSION_STR},"
                f" but version {torch.__version__} was found instead."
            ),
        )


except ORTModuleFallbackException as e:
    # Initialization fallback is handled at ORTModule.__init__
    _FALLBACK_INIT_EXCEPTION = e
except ImportError as e:
    raise RuntimeError(
        f"PyTorch {MINIMUM_RUNTIME_PYTORCH_VERSION_STR} must be "
        "installed in order to run ONNX Runtime ORTModule frontend!"
    ) from e

# Verify whether PyTorch C++ extensions are already compiled
# TODO: detect when installed extensions are outdated and need reinstallation. Hash? Version file?
if not is_torch_cpp_extensions_installed(ORTMODULE_TORCH_CPP_DIR) and "-m" not in sys.argv:
    _FALLBACK_INIT_EXCEPTION = wrap_exception(
        ORTModuleInitException,
        RuntimeError(
            f"ORTModule's extensions were not detected at '{ORTMODULE_TORCH_CPP_DIR}' folder. "
            "Run `python -m torch_ort.configure` before using `ORTModule` frontend."
        ),
    )

# Initialized ORT's random seed with pytorch's initial seed
# in case user has set pytorch seed before importing ORTModule
set_seed(torch.initial_seed() % sys.maxsize)


# Override torch.manual_seed and torch.cuda.manual_seed
def override_torch_manual_seed(seed):
    set_seed(int(seed % sys.maxsize))
    return torch_manual_seed(seed)


torch_manual_seed = torch.manual_seed
torch.manual_seed = override_torch_manual_seed


def override_torch_cuda_manual_seed(seed):
    set_seed(int(seed % sys.maxsize))
    return torch_cuda_manual_seed(seed)


torch_cuda_manual_seed = torch.cuda.manual_seed
torch.cuda.manual_seed = override_torch_cuda_manual_seed


def _use_deterministic_algorithms(enabled):
    global ORTMODULE_IS_DETERMINISTIC  # noqa: PLW0603
    ORTMODULE_IS_DETERMINISTIC = enabled


def _are_deterministic_algorithms_enabled():
    global ORTMODULE_IS_DETERMINISTIC  # noqa: PLW0602
    return ORTMODULE_IS_DETERMINISTIC


from .graph_optimizer_registry import register_graph_optimizer  # noqa: E402, F401
from .graph_optimizers import *  # noqa: E402, F403
from .options import DebugOptions, LogLevel  # noqa: E402, F401

# ORTModule must be loaded only after all validation passes
from .ortmodule import ORTModule  # noqa: E402, F401



# --------------------------------------------------------------------------
# Copyright 2020 The HuggingFace Inc. team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import os
import random
import unittest
from pathlib import Path

import numpy
import onnx
import pytest
import torch
from onnx import helper
from parity_utilities import compare_outputs, create_ort_session, find_transformers_source, parse_arguments
from torch import nn
from transformers.modeling_utils import Conv1D

if find_transformers_source():
    from onnx_model import OnnxModel
else:
    from onnxruntime.transformers.onnx_model import OnnxModel

DEBUG_OUTPUTS = ["qk", "norm_qk", "softmax", "attn_weights"]


class MyGPT2Attention(nn.Module):
    """
    This module is modifed from Gpt2Attention of huggingface transformers v4.9.1.
    Code related to crosss attention, c_proj, attn_dropout and head_mask etc are removed.
    """

    def __init__(
        self,
        max_position_embeddings=1024,
        hidden_size=768,
        num_attention_heads=12,
        use_cache=True,
        debug=False,
        fix_onnx_export=True,
    ):
        super().__init__()
        max_positions = max_position_embeddings
        self.register_buffer(
            "bias",
            torch.tril(torch.ones((max_positions, max_positions), dtype=torch.uint8)).view(
                1, 1, max_positions, max_positions
            ),
        )
        self.register_buffer("masked_bias", torch.tensor(-1e4))
        self.embed_dim = hidden_size
        self.num_heads = num_attention_heads
        self.head_dim = self.embed_dim // self.num_heads
        self.split_size = self.embed_dim
        assert self.head_dim * self.num_heads == self.embed_dim

        self.c_attn = Conv1D(3 * self.embed_dim, self.embed_dim)
        # Use random bias instead of zeros for parity test.
        self.c_attn.bias = nn.Parameter(torch.normal(0.0, 0.1, (3 * self.embed_dim,)))

        self.use_cache = use_cache
        self.debug = debug
        self.fix_onnx_export = fix_onnx_export

    def _attn(self, query, key, value, attention_mask=None):
        qk = torch.matmul(query, key.transpose(-1, -2))

        # Torch has special handling for Div and Mul by a scalar:
        #   https://github.com/pytorch/pytorch/blob/5536cda19a5def9e0553b318f04d297d602ac956/aten/src/ATen/native/cuda/BinaryMulDivKernel.cu#L52-L60
        #   https://github.com/pytorch/pytorch/blob/5536cda19a5def9e0553b318f04d297d602ac956/aten/src/ATen/native/cuda/BinaryMulDivKernel.cu#L185-L194
        # Modify the code to use same processing in onnx export so as to get parity result without attention fusion.
        # This walkaround is not needed when attention fusion will be applied later since the subgraph will be replaced by an Attention node.
        if self.fix_onnx_export and torch.onnx.is_in_onnx_export():
            if qk.dtype == torch.float16:
                norm_qk = qk.to(torch.float32) * (1.0 / (float(value.size(-1)) ** 0.5))
                norm_qk = norm_qk.to(torch.float16)
            else:
                norm_qk = qk * (1.0 / (float(value.size(-1)) ** 0.5))
        else:
            norm_qk = qk / (float(value.size(-1)) ** 0.5)

        query_length, key_length = query.size(-2), key.size(-2)
        causal_mask = self.bias[:, :, key_length - query_length : key_length, :key_length].bool()
        attn_weights = torch.where(causal_mask, norm_qk, self.masked_bias.to(norm_qk.dtype))

        if attention_mask is not None:
            attn_weights = attn_weights + attention_mask

        softmax = nn.Softmax(dim=-1)(attn_weights)

        attn_output = torch.matmul(softmax, value)

        if self.debug:
            return attn_output, qk, norm_qk, softmax, attn_weights
        else:
            return attn_output

    def _split_heads(self, tensor, num_heads, attn_head_size):
        new_shape = tensor.size()[:-1] + (num_heads, attn_head_size)
        tensor = tensor.view(*new_shape)
        return tensor.permute(0, 2, 1, 3)  # (batch, head, seq_length, head_features)

    def _merge_heads(self, tensor, num_heads, attn_head_size):
        tensor = tensor.permute(0, 2, 1, 3).contiguous()
        new_shape = tensor.size()[:-2] + (num_heads * attn_head_size,)
        return tensor.view(new_shape)

    @staticmethod
    def concat_key_value(key, value):
        return torch.cat((key.unsqueeze(0), value.unsqueeze(0)), dim=0)

    @staticmethod
    def process_mask(attention_mask, dtype):
        # Create a 4D attention mask with shape [batch_size, 1, 1, to_seq_length] from a 2D tensor mask.
        attention_mask = attention_mask[:, None, None, :]
        attention_mask = attention_mask.to(dtype=dtype)  # fp16 compatibility
        attention_mask = (1.0 - attention_mask) * -10000.0
        return attention_mask

    def forward(self, hidden_states, attention_mask=None, layer_past=None):
        query, key, value = self.c_attn(hidden_states).split(self.split_size, dim=2)

        query = self._split_heads(query, self.num_heads, self.head_dim)
        key = self._split_heads(key, self.num_heads, self.head_dim)
        value = self._split_heads(value, self.num_heads, self.head_dim)

        if layer_past is not None:
            past_key, past_value = layer_past
            key = torch.cat((past_key, key), dim=-2)
            value = torch.cat((past_value, value), dim=-2)

        if self.use_cache is True:
            # Instead of present = (key, value), here we merge them into one tensor to be compatible with Attention operator.
            present = MyGPT2Attention.concat_key_value(key, value)
        else:
            present = None

        mask = MyGPT2Attention.process_mask(attention_mask, dtype=query.dtype)  # mask processing is moved to here.

        if self.debug:
            attn_output, qk, norm_qk, softmax, attn_weights = self._attn(query, key, value, mask)
        else:
            attn_output = self._attn(query, key, value, mask)

        attn_output = self._merge_heads(attn_output, self.num_heads, self.head_dim)

        outputs = (attn_output, present)
        if self.debug:
            if "qk" in DEBUG_OUTPUTS:
                outputs += (qk,)
            if "norm_qk" in DEBUG_OUTPUTS:
                outputs += (norm_qk,)
            if "softmax" in DEBUG_OUTPUTS:
                outputs += (softmax,)
            if "attn_weights" in DEBUG_OUTPUTS:
                outputs += (attn_weights,)

        return outputs


def create_inputs(
    batch_size=1,
    hidden_size=768,
    num_attention_heads=12,
    sequence_length=1,
    past_sequence_length=5,
    float16=False,
    device=torch.device("cuda"),  # noqa: B008
    padding_length=0,
):
    float_type = torch.float16 if float16 else torch.float32

    past_shape = [
        batch_size,
        num_attention_heads,
        past_sequence_length,
        int(hidden_size / num_attention_heads),
    ]
    past_key = torch.rand(past_shape, dtype=float_type, device=device)
    past_value = torch.rand(past_shape, dtype=float_type, device=device)
    layer_past = MyGPT2Attention.concat_key_value(past_key, past_value)

    total_sequence_length = past_sequence_length + sequence_length

    attention_mask = torch.ones([batch_size, total_sequence_length], dtype=torch.int32, device=device)
    if padding_length > 0:
        padding_position = total_sequence_length - padding_length
        attention_mask[:, padding_position:] = 0
    elif padding_length < 0:  # mask a random position
        for i in range(batch_size):
            padding_position = random.randint(0, total_sequence_length - 1)
            attention_mask[i, padding_position] = 0

    input_hidden_states = (
        torch.normal(mean=0.0, std=0.1, size=(batch_size, sequence_length, hidden_size)).to(float_type).to(device)
    )
    return input_hidden_states, attention_mask, layer_past


def get_output_names(debug=False):
    outputs = ["attn_output", "present"]
    if debug:
        outputs += DEBUG_OUTPUTS
    return outputs


def export_onnx(model, onnx_model_path, float16, hidden_size, num_attention_heads, debug, device):
    Path(onnx_model_path).parent.mkdir(parents=True, exist_ok=True)

    input_hidden_states, attention_mask, layer_past = create_inputs(
        float16=float16,
        hidden_size=hidden_size,
        num_attention_heads=num_attention_heads,
        device=device,
    )

    with torch.no_grad():
        model(input_hidden_states, attention_mask=attention_mask, layer_past=layer_past)

    dynamic_axes = {
        "input_hidden_states": {0: "batch_size", 1: "seq_len"},
        "attn_output": {0: "batch_size", 1: "seq_len"},
        "past": {1: "batch_size", 3: "past_seq_len"},
        "present": {1: "batch_size", 3: "total_seq_len"},
        "attention_mask": {0: "batch_size", 1: "total_seq_len"},
    }
    if debug:
        debug_dynamic_axes = {
            "qk": {0: "batch_size", 1: "seq_len"},
            "norm_qk": {0: "batch_size", 1: "seq_len"},
            "softmax": {0: "batch_size", 1: "seq_len"},
            "attn_weights": {0: "batch_size", 1: "seq_len"},
        }
        for name in DEBUG_OUTPUTS:
            dynamic_axes[name] = debug_dynamic_axes[name]

    torch.onnx.export(
        model,
        args=(
            input_hidden_states,
            {"attention_mask": attention_mask, "layer_past": layer_past},
        ),
        f=onnx_model_path,
        input_names=["input_hidden_states", "attention_mask", "past"],
        output_names=get_output_names(debug),
        dynamic_axes=dynamic_axes,
        opset_version=11,
        do_constant_folding=True,
    )
    print("exported:", onnx_model_path)


def optimize_onnx(input_onnx_path, optimized_onnx_path, num_heads, debug):
    m = onnx.load(input_onnx_path)
    onnx_model = OnnxModel(m)

    nodes_to_remove = onnx_model.nodes()
    output_names = ["attn_output", "present", *DEBUG_OUTPUTS] if debug else ["attn_output", "present"]
    node_to_add = helper.make_node(
        "Attention",
        [
            "input_hidden_states",
            "c_attn.weight",
            "c_attn.bias",
            "attention_mask",
            "past",
        ],
        output_names,
        "gpt2_attention",
        num_heads=num_heads,
        unidirectional=1,
        domain="com.microsoft",
    )

    onnx_model.remove_nodes(nodes_to_remove)
    onnx_model.add_node(node_to_add)
    onnx_model.prune_graph()
    onnx_model.save_model_to_file(optimized_onnx_path)


def onnxruntime_inference(ort_session, input_hidden_states, attention_mask, past):
    ort_inputs = {
        "past": numpy.ascontiguousarray(past.cpu().numpy()),
        "attention_mask": numpy.ascontiguousarray(attention_mask.cpu().numpy()),
        "input_hidden_states": numpy.ascontiguousarray(input_hidden_states.cpu().numpy()),
    }

    ort_outputs = ort_session.run(None, ort_inputs)
    return ort_outputs


def verify_attention(
    model,
    onnx_model_path,
    batch_size,
    hidden_size,
    num_attention_heads,
    sequence_length,
    past_sequence_length,
    float16,
    device,
    padding_length,
    optimized,
    test_cases=100,
    verbose=False,
):
    print(
        f"optimized={optimized}, batch_size={batch_size}, hidden_size={hidden_size}, num_attention_heads={num_attention_heads}, sequence_length={sequence_length}, past_sequence_length={past_sequence_length}, float16={float16}, padding_length={padding_length}, device={device}"
    )
    passed_cases = 0
    max_diffs = []

    ort_session = create_ort_session(onnx_model_path, device.type == "cuda", verbose=verbose)
    for _i in range(test_cases):
        input_hidden_states, attention_mask, layer_past = create_inputs(
            batch_size,
            hidden_size,
            num_attention_heads,
            sequence_length,
            past_sequence_length,
            float16,
            device,
            padding_length,
        )

        with torch.no_grad():
            torch_outputs = model(
                input_hidden_states,
                layer_past=layer_past,
                attention_mask=attention_mask,
            )

        ort_outputs = onnxruntime_inference(ort_session, input_hidden_states, attention_mask, layer_past)

        tolerance = 1e-02 if float16 else 1e-04
        is_all_close, max_diff = compare_outputs(torch_outputs, ort_outputs, atol=tolerance, verbose=True)
        max_diffs.append(max_diff)
        if is_all_close:
            passed_cases += 1

    max_diff = max(max_diffs)
    diff_count = len([i for i in max_diffs if i > 0])
    success_flag = "[FAILED]" if passed_cases < test_cases else "[OK]"
    print(f"{success_flag} Passed_cases={passed_cases}/{test_cases}; Max_diff={max_diff}; Diff_count={diff_count}")
    return test_cases - passed_cases


def run(batch_size, float16, optimized, hidden_size, num_attention_heads, device, test_cases, verbose=False):
    test_name = f"batch_size={batch_size}, float16={float16}, optimized={optimized}, hidden_size={hidden_size}, num_attention_heads={num_attention_heads}"
    print(f"\nTesting ONNX parity: {test_name}")

    debug = (
        not optimized
    )  # or DEBUG_OUTPUTS==["softmax"] when you add an extra output for softmax result in Attention operator
    model = MyGPT2Attention(hidden_size=hidden_size, num_attention_heads=num_attention_heads, debug=debug)
    model.eval()
    model.to(device)
    if float16:
        model.half()

    # Do not re-use onnx file from previous test since weights of model are random.
    onnx_model_path = "./temp/gpt_attention_{}.onnx".format("fp16" if float16 else "fp32")
    export_onnx(model, onnx_model_path, float16, hidden_size, num_attention_heads, debug, device)

    if optimized:
        optimized_onnx_path = "./temp/gpt_attention_opt_{}.onnx".format("fp16" if float16 else "fp32")
        optimize_onnx(onnx_model_path, optimized_onnx_path, num_attention_heads, debug)
        onnx_path = optimized_onnx_path
    else:
        onnx_path = onnx_model_path

    # Test Case: No past state
    sequence_length = 2
    past_sequence_length = 0
    padding_length = 0
    num_failure = 0
    num_failure += verify_attention(
        model,
        onnx_path,
        batch_size,
        hidden_size,
        num_attention_heads,
        sequence_length,
        past_sequence_length,
        float16,
        device,
        padding_length,
        optimized,
        test_cases,
        verbose,
    )

    # Test Case: with past state and padding last 2 words
    sequence_length = 3
    past_sequence_length = 5
    padding_length = 2
    num_failure += verify_attention(
        model,
        onnx_path,
        batch_size,
        hidden_size,
        num_attention_heads,
        sequence_length,
        past_sequence_length,
        float16,
        device,
        padding_length,
        optimized,
        test_cases,
        verbose,
    )

    # Test Case: random mask one word
    sequence_length = 1
    past_sequence_length = 128
    padding_length = -1
    num_failure += verify_attention(
        model,
        onnx_path,
        batch_size,
        hidden_size,
        num_attention_heads,
        sequence_length,
        past_sequence_length,
        float16,
        device,
        padding_length,
        optimized,
        test_cases,
        verbose,
    )

    # clean up onnx file
    os.remove(onnx_model_path)
    if optimized:
        os.remove(onnx_path)

    return num_failure, test_name


class TestGptAttentionHuggingfaceParity(unittest.TestCase):
    verbose = False
    optimized = True

    def setUp(self):
        self.test_cases = 10  # Number of test cases per test run

    def run_test(self, batch_size, float16, optimized, hidden_size, num_attention_heads, device, verbose=False):
        if float16 and device.type == "cpu":  # CPU does not support FP16
            return
        num_failure, test_name = run(
            batch_size,
            float16,
            optimized,
            hidden_size,
            num_attention_heads,
            device,
            self.test_cases,
            verbose=verbose,
        )
        self.assertTrue(num_failure == 0, test_name)

    def run_small(self, optimized, device, verbose=False):
        for batch_size in [64]:
            self.run_test(
                batch_size,
                float16=False,
                optimized=optimized,
                hidden_size=768,
                num_attention_heads=12,
                device=device,
                verbose=verbose,
            )
            self.run_test(
                batch_size,
                float16=True,
                optimized=optimized,
                hidden_size=768,
                num_attention_heads=12,
                device=device,
                verbose=verbose,
            )

    def run_large(self, optimized, device, verbose=False):
        for batch_size in [2]:
            self.run_test(
                batch_size,
                float16=False,
                optimized=optimized,
                hidden_size=4096,
                num_attention_heads=32,
                device=device,
                verbose=verbose,
            )
            self.run_test(
                batch_size,
                float16=True,
                optimized=optimized,
                hidden_size=4096,
                num_attention_heads=32,
                device=device,
                verbose=verbose,
            )

    def test_cpu(self):
        cpu = torch.device("cpu")
        self.run_small(self.optimized, cpu, verbose=self.verbose)

    def test_cuda(self):
        if not torch.cuda.is_available():
            self.skipTest("test requires GPU and torch+cuda")
        else:
            gpu = torch.device("cuda")
            self.run_small(self.optimized, gpu, verbose=self.verbose)

    @pytest.mark.slow
    def test_large_cuda(self):
        if not torch.cuda.is_available():
            self.skipTest("test requires GPU and torch+cuda")
        else:
            gpu = torch.device("cuda")
            self.run_large(self.optimized, gpu, verbose=self.verbose)


if __name__ == "__main__":
    args, remaining_args = parse_arguments(namespace_filter=unittest)

    TestGptAttentionHuggingfaceParity.verbose = args.log_verbose
    TestGptAttentionHuggingfaceParity.optimized = args.optimize

    unittest.main(argv=remaining_args)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import os
import tempfile

import numpy as np
import onnx
import pytest
from onnx import TensorProto, helper

import onnxruntime as ort


def sigmoid(z):
    """Computes the sigmoid of the given numpy array."""
    return 1 / (1 + np.exp(-z))


class LSTM:
    """LSTM utility class for testing.

    This class exposes four copmutation methods:
    - forward_np: computes the LSTM forward pass using numpy
    - forward_ort: computes the LSTM forward pass using ORT
    - backward_np: computes the LSTM backward pass using numpy
    - backward_ort: computes the LSTM backward pass using ORT
    and two onnx model generation methods:
    - forward_graph: generates the LSTM forward onnx graph (with the LSTMTraining node)
    - backward_graph: generates the LSTM backward onnx graph (with the LSTMGrad node)
    """

    def __init__(self, sequence_length, batch_size, input_size, hidden_size):
        """Initializes the LSTM class.

        Args:
            sequence_length (int): the sequence length
            batch_size (int): the batch size
            input_size (int): the input size
            hidden_size (int): the hidden size
        """
        self._sequence_length = sequence_length
        self._batch_size = batch_size
        self.input_size = input_size
        self._hidden_size = hidden_size
        self._num_directions = 1
        self._forward_model = None
        self._backward_model = None

    def forward_np(
        self,
        inputs,
        weights,
        recurrence_weights,
        bias=None,
        initial_hidden_state=None,
        initial_cell_state=None,
        peephole_weights=None,
    ):
        """Computes the LSTM forward pass using numpy.

        The computation follows the following rules:
        - it = sigmoid(Xt*(Wi^T) + Ht-1*(Ri^T) + Pi (.) Ct-1 + Wbi + Rbi)
        - ft = sigmoid(Xt*(Wf^T) + Ht-1*(Rf^T) + Pf (.) Ct-1 + Wbf + Rbf)
        - ct = tanh(Xt*(Wc^T) + Ht-1*(Rc^T) + Wbc + Rbc)
        - Ct = ft (.) Ct-1 + it (.) ct
        - ot = sigmoid(Xt*(Wo^T) + Ht-1*(Ro^T) + Po (.) Ct + Wbo + Rbo)
        - Ht = ot (.) h(Ct)


        Args:
            input (np.array): the input tensor of shape (sequence_length, batch_size, input_size)
            weights (np.array): the weight tensor of shape
                                (num_directions, 4 * hidden_size, input_size)
            recurrence_weights (np.array): the recurrence weight tensor of shape
                                           (num_directions, 4 * hidden_size, hidden_size)
            bias (np.array, optional): the bias tensor of shape
                                       (num_directions, 8 * hidden_size). Defaults to None.
            H0 (np.array, optional): the initial hidden state tensor of shape
                                     (num_directions, batch_size, hidden_size).
                                     Defaults to None.
            C0 (np.array, optional): the initial cell state tensor of shape
                                     (num_directions, batch_size, hidden_size).
                                     Defaults to None.
            P (np.array, optional): the peephole weight tensor of shape
                                    (num_directions, 3 * hidden_size).
                                    Defaults to None.

        Returns:
            HAll (np.array): all hidden states tensor of shape
                             (sequence_length, num_directions, batch_size, hidden_size)
            HFinal (np.array): the final hidden state tensor of shape
                               (num_directions, batch_size, hidden_size)
            CFinal (np.array): the final cell state tensor of shape
                               (num_directions, batch_size, hidden_size)
            CAll (np.array): all cell states tensor of shape
                             (sequence_length, num_directions, batch_size, hidden_size)
            IOFC (np.array): all intermediate values of the gates tensor of shape
                             (sequence_length, num_directions, batch_size, 4 * hidden_size)
        """
        all_hidden_states = np.zeros(
            (self._sequence_length, self._num_directions, self._batch_size, self._hidden_size), np.float32
        )
        final_hidden_state = np.zeros((self._num_directions, self._batch_size, self._hidden_size), np.float32)
        final_cell_state = np.zeros((self._num_directions, self._batch_size, self._hidden_size), np.float32)
        all_cell_states = np.zeros(
            (self._sequence_length, self._num_directions, self._batch_size, self._hidden_size), np.float32
        )
        iofc = np.zeros(
            (self._sequence_length, self._num_directions, self._batch_size, 4 * self._hidden_size), np.float32
        )
        for idx in range(self._batch_size):
            prev_h = (
                initial_hidden_state[0, idx, :]
                if initial_hidden_state is not None
                else np.zeros((self._hidden_size), np.float32)
            )
            prev_c = (
                initial_cell_state[0, idx, :]
                if initial_cell_state is not None
                else np.zeros((self._hidden_size), np.float32)
            )
            for t in range(self._sequence_length):
                current_input = inputs[t, idx, :]
                weights_i = np.squeeze(weights[:, : self._hidden_size, :], axis=0)
                weights_o = np.squeeze(weights[:, self._hidden_size : 2 * self._hidden_size, :], axis=0)
                weights_f = np.squeeze(weights[:, 2 * self._hidden_size : 3 * self._hidden_size, :], axis=0)
                weights_c = np.squeeze(weights[:, 3 * self._hidden_size :, :], axis=0)

                rweights_i = np.squeeze(recurrence_weights[:, : self._hidden_size, :], axis=0)
                rweights_o = np.squeeze(recurrence_weights[:, self._hidden_size : 2 * self._hidden_size, :], axis=0)
                rweights_f = np.squeeze(recurrence_weights[:, 2 * self._hidden_size : 3 * self._hidden_size, :], axis=0)
                rweights_c = np.squeeze(recurrence_weights[:, 3 * self._hidden_size :, :], axis=0)

                wb_i = (
                    np.squeeze(bias[:, : self._hidden_size], axis=0)
                    if bias is not None
                    else np.zeros((self._hidden_size), np.float32)
                )
                wb_o = (
                    np.squeeze(bias[:, self._hidden_size : 2 * self._hidden_size], axis=0)
                    if bias is not None
                    else np.zeros((self._hidden_size), np.float32)
                )
                wb_f = (
                    np.squeeze(bias[:, 2 * self._hidden_size : 3 * self._hidden_size], axis=0)
                    if bias is not None
                    else np.zeros((self._hidden_size), np.float32)
                )
                wb_c = (
                    np.squeeze(bias[:, 3 * self._hidden_size : 4 * self._hidden_size], axis=0)
                    if bias is not None
                    else np.zeros((self._hidden_size), np.float32)
                )
                rb_i = (
                    np.squeeze(bias[:, 4 * self._hidden_size : 5 * self._hidden_size], axis=0)
                    if bias is not None
                    else np.zeros((self._hidden_size), np.float32)
                )
                rb_o = (
                    np.squeeze(bias[:, 5 * self._hidden_size : 6 * self._hidden_size], axis=0)
                    if bias is not None
                    else np.zeros((self._hidden_size), np.float32)
                )
                rb_f = (
                    np.squeeze(bias[:, 6 * self._hidden_size : 7 * self._hidden_size], axis=0)
                    if bias is not None
                    else np.zeros((self._hidden_size), np.float32)
                )
                rb_c = (
                    np.squeeze(bias[:, 7 * self._hidden_size :], axis=0)
                    if bias is not None
                    else np.zeros((self._hidden_size), np.float32)
                )

                peephole_weights_i = (
                    np.squeeze(peephole_weights[:, : self._hidden_size], axis=0)
                    if peephole_weights is not None
                    else np.zeros((self._hidden_size), np.float32)
                )
                peephole_weights_o = (
                    np.squeeze(peephole_weights[:, self._hidden_size : 2 * self._hidden_size], axis=0)
                    if peephole_weights is not None
                    else np.zeros((self._hidden_size), np.float32)
                )
                peephole_weights_f = (
                    np.squeeze(peephole_weights[:, 2 * self._hidden_size : 3 * self._hidden_size], axis=0)
                    if peephole_weights is not None
                    else np.zeros((self._hidden_size), np.float32)
                )

                input_gate = sigmoid(
                    np.dot(current_input, weights_i.T)
                    + np.dot(prev_h, rweights_i.T)
                    + wb_i
                    + rb_i
                    + peephole_weights_i * prev_c
                )
                forget_gate = sigmoid(
                    np.dot(current_input, weights_f.T)
                    + np.dot(prev_h, rweights_f.T)
                    + wb_f
                    + rb_f
                    + peephole_weights_f * prev_c
                )
                control_gate = np.tanh(np.dot(current_input, weights_c.T) + np.dot(prev_h, rweights_c.T) + wb_c + rb_c)

                final_cell_state[0, idx, :] = forget_gate * prev_c + input_gate * control_gate

                output_gate = sigmoid(
                    np.dot(current_input, weights_o.T)
                    + np.dot(prev_h, rweights_o.T)
                    + wb_o
                    + rb_o
                    + peephole_weights_o * final_cell_state[0, idx, :]
                )

                iofc[t, 0, idx, : self._hidden_size] = input_gate
                iofc[t, 0, idx, self._hidden_size : 2 * self._hidden_size] = output_gate
                iofc[t, 0, idx, 2 * self._hidden_size : 3 * self._hidden_size] = forget_gate
                iofc[t, 0, idx, 3 * self._hidden_size :] = control_gate

                final_hidden_state[0, idx, :] = output_gate * np.tanh(final_cell_state[0, idx, :])

                all_hidden_states[t, 0, idx, :] = final_hidden_state[0, idx, :]
                all_cell_states[t, 0, idx, :] = final_cell_state[0, idx, :]

                prev_h = final_hidden_state[0, idx, :]
                prev_c = final_cell_state[0, idx, :]

        return all_hidden_states, final_hidden_state, final_cell_state, all_cell_states, iofc

    def forward_ort(
        self,
        inputs,
        weights,
        recurrence_weights,
        bias=None,
        initial_hidden_state=None,
        initial_cell_state=None,
        peephole_weights=None,
    ):
        """Run LSTM forward pass using ONNX Runtime."""
        ort_inputs = {"inputs": inputs, "weights": weights, "recurrence_weights": recurrence_weights}
        if bias is not None:
            ort_inputs["bias"] = bias
        if initial_hidden_state is not None:
            ort_inputs["initial_hidden_state"] = initial_hidden_state
        if initial_cell_state is not None:
            ort_inputs["initial_cell_state"] = initial_cell_state
        if peephole_weights is not None:
            ort_inputs["peephole_weights"] = peephole_weights

        ort_outs = None
        with tempfile.TemporaryDirectory() as tmpdirname:
            model_path = os.path.join(tmpdirname, "lstm.onnx")
            onnx.save(self._forward_model, model_path)
            ort_session = ort.InferenceSession(model_path, providers=["CPUExecutionProvider"])
            ort_outs = ort_session.run(None, ort_inputs)

        return ort_outs

    def forward_graph(
        self,
        bias=True,
        sequence_lengths=False,
        initial_hidden_state=True,
        initial_cell_state=True,
        peephole_weights=False,
    ):
        """Create a graph for LSTM forward pass."""
        inputs = helper.make_tensor_value_info(
            "inputs", TensorProto.FLOAT, [self._sequence_length, self._batch_size, self.input_size]
        )
        weights = helper.make_tensor_value_info(
            "weights", TensorProto.FLOAT, [self._num_directions, 4 * self._hidden_size, self.input_size]
        )
        recurrence_weights = helper.make_tensor_value_info(
            "recurrence_weights", TensorProto.FLOAT, [self._num_directions, 4 * self._hidden_size, self._hidden_size]
        )
        bias = (
            helper.make_tensor_value_info("bias", TensorProto.FLOAT, [self._num_directions, 8 * self._hidden_size])
            if bias
            else None
        )
        sequence_lengths = (
            helper.make_tensor_value_info("sequence_lengths", TensorProto.INT64, [self._batch_size])
            if sequence_lengths
            else None
        )
        initial_hidden_state = (
            helper.make_tensor_value_info(
                "initial_hidden_state", TensorProto.FLOAT, [self._num_directions, self._batch_size, self._hidden_size]
            )
            if initial_hidden_state
            else None
        )
        initial_cell_state = (
            helper.make_tensor_value_info(
                "initial_cell_state", TensorProto.FLOAT, [self._num_directions, self._batch_size, self._hidden_size]
            )
            if initial_cell_state
            else None
        )
        peephole_weights = (
            helper.make_tensor_value_info(
                "peephole_weights", TensorProto.FLOAT, [self._num_directions, 3 * self._hidden_size]
            )
            if peephole_weights
            else None
        )

        all_hidden_states = helper.make_tensor_value_info(
            "all_hidden_states",
            TensorProto.FLOAT,
            [self._sequence_length, self._num_directions, self._batch_size, self._hidden_size],
        )
        final_hidden_state = helper.make_tensor_value_info(
            "final_hidden_state", TensorProto.FLOAT, [self._num_directions, self._batch_size, self._hidden_size]
        )
        final_cell_state = helper.make_tensor_value_info(
            "final_cell_state", TensorProto.FLOAT, [self._num_directions, self._batch_size, self._hidden_size]
        )
        all_cell_states = helper.make_tensor_value_info(
            "all_cell_states",
            TensorProto.FLOAT,
            [self._sequence_length, self._num_directions, self._batch_size, self._hidden_size],
        )
        iofc = helper.make_tensor_value_info(
            "iofc",
            TensorProto.FLOAT,
            [self._sequence_length, self._num_directions, self._batch_size, 4 * self._hidden_size],
        )

        lstm = helper.make_node(
            "LSTMTraining",
            inputs=[
                "inputs",
                "weights",
                "recurrence_weights",
                "bias" if bias else "",
                "sequence_lengths" if sequence_lengths else "",
                "initial_hidden_state" if initial_hidden_state else "",
                "initial_cell_state" if initial_cell_state else "",
                "peephole_weights" if peephole_weights else "",
            ],
            outputs=["all_hidden_states", "final_hidden_state", "final_cell_state", "all_cell_states", "iofc"],
            domain="com.microsoft",
            hidden_size=self._hidden_size,
        )

        graph = helper.make_graph(
            [lstm],
            "lstm",
            [
                gi
                for gi in [
                    inputs,
                    weights,
                    recurrence_weights,
                    bias,
                    sequence_lengths,
                    initial_hidden_state,
                    initial_cell_state,
                    peephole_weights,
                ]
                if gi
            ],
            [all_hidden_states, final_hidden_state, final_cell_state, all_cell_states, iofc],
        )

        self._forward_model = helper.make_model(
            graph,
            producer_name="orttraining",
            opset_imports=[helper.make_opsetid("", 14), helper.make_opsetid("com.microsoft", 1)],
        )

        return self._forward_model

    def backward_np(
        self,
        inputs,
        weights,
        recurrence_weights,
        bias,
        initial_hidden_state,
        initial_cell_state,
        peephole_weights,
        all_hidden_states,
        all_cell_states,
        iofc,
        grad_all_hidden_states,
        grad_final_hidden_state=None,
        grad_final_cell_state=None,
    ):
        """Compute the backward pass of LSTM using numpy.

        This is a reference implementation for testing purpose. The equations used here are from
        deriving the gradients based on:
        - it = sigmoid(Xt*(Wi^T) + Ht-1*(Ri^T) + Pi (.) Ct-1 + Wbi + Rbi)
        - ft = sigmoid(Xt*(Wf^T) + Ht-1*(Rf^T) + Pf (.) Ct-1 + Wbf + Rbf)
        - ct = tanh(Xt*(Wc^T) + Ht-1*(Rc^T) + Wbc + Rbc)
        - Ct = ft (.) Ct-1 + it (.) ct
        - ot = sigmoid(Xt*(Wo^T) + Ht-1*(Ro^T) + Po (.) Ct + Wbo + Rbo)
        - Ht = ot (.) h(Ct)


        Args:
            inputs (np.ndarray): input tensor of shape (sequence_length, batch_size, input_size)
            weights (np.ndarray): weight tensor of shape (num_directions, 4 * hidden_size, input_size)
            recurrence_weights (np.ndarray): recurrence weight tensor of shape (num_directions, 4 * hidden_size, hidden_size)
            bias (bool): whether to compute the bias gradient or not
            initial_hidden_state (np.ndarray): initial hidden state tensor of shape (num_directions, batch_size, hidden_size)
            initial_cell_state (np.ndarray): initial cell state tensor of shape (num_directions, batch_size, hidden_size)
            peephole_weights (bool): whether to compute the peephole weights gradient or not
            all_hidden_states (np.ndarray): output tensor of shape (sequence_length, num_directions, batch_size, hidden_size)
            all_cell_states (np.ndarray): cell state tensor of shape (sequence_length, num_directions, batch_size, hidden_size)
            iofc (np.ndarray): input, output, forget, and cell gate tensor of shape (sequence_length, num_directions, batch_size, 4 * hidden_size)
            grad_all_hidden_states (np.ndarray): gradient of HAll
            grad_final_hidden_state (np.ndarray): gradient of Ht
            grad_final_cell_state (np.ndarray): gradient of Ct

        Returns:
            tuple[np.ndarray]: gradients of inputs, weights, recurrence_weights, bias, initial_hidden_state, initial_cell_state, peephole_weightsP
        """
        grad_inputs = np.zeros((self._sequence_length, self._batch_size, self.input_size), np.float32)
        grad_weights = np.zeros((self._num_directions, 4 * self._hidden_size, self.input_size), np.float32)
        grad_recurrence_weights = np.zeros((self._num_directions, 4 * self._hidden_size, self._hidden_size), np.float32)
        grad_bias = np.zeros((self._num_directions, 8 * self._hidden_size), np.float32) if bias is True else None
        grad_initial_hidden_state = (
            np.zeros((self._num_directions, self._batch_size, self._hidden_size), np.float32)
            if initial_hidden_state is not None
            else None
        )
        grad_initial_cell_state = (
            np.zeros((self._num_directions, self._batch_size, self._hidden_size), np.float32)
            if initial_cell_state is not None
            else None
        )
        grad_peephole_weights = (
            np.zeros((self._num_directions, 3 * self._hidden_size), np.float32) if peephole_weights is True else None
        )

        for idx in range(self._batch_size):
            grad_h = (
                grad_final_hidden_state[0, idx, :]
                if grad_final_hidden_state is not None
                else np.zeros((self._hidden_size), np.float32)
            )
            grad_c = (
                grad_final_cell_state[0, idx, :]
                if grad_final_cell_state is not None
                else np.zeros((self._hidden_size), np.float32)
            )
            for t in reversed(range(self._sequence_length)):
                current_input = inputs[t, idx, :]
                weights_i = np.squeeze(weights[:, : self._hidden_size, :], axis=0)
                weights_o = np.squeeze(weights[:, self._hidden_size : 2 * self._hidden_size, :], axis=0)
                weights_f = np.squeeze(weights[:, 2 * self._hidden_size : 3 * self._hidden_size, :], axis=0)
                weights_c = np.squeeze(weights[:, 3 * self._hidden_size :, :], axis=0)

                rweights_i = np.squeeze(recurrence_weights[:, : self._hidden_size, :], axis=0)
                rweights_o = np.squeeze(recurrence_weights[:, self._hidden_size : 2 * self._hidden_size, :], axis=0)
                rweights_f = np.squeeze(recurrence_weights[:, 2 * self._hidden_size : 3 * self._hidden_size, :], axis=0)
                rweights_c = np.squeeze(recurrence_weights[:, 3 * self._hidden_size :, :], axis=0)

                input_gate = iofc[t, 0, idx, : self._hidden_size]
                output_gate = iofc[t, 0, idx, self._hidden_size : 2 * self._hidden_size]
                forget_gate = iofc[t, 0, idx, 2 * self._hidden_size : 3 * self._hidden_size]
                control_gate = iofc[t, 0, idx, 3 * self._hidden_size :]

                grad_h += grad_all_hidden_states[t, 0, idx, :]
                grad_c += grad_h * output_gate * (1 - np.tanh(all_cell_states[t, 0, idx, :]) ** 2)
                grad_prev_c = grad_c * forget_gate
                if t == 0 and grad_initial_cell_state is not None:
                    grad_initial_cell_state[0, idx, :] = grad_prev_c

                grad_input_gate = grad_c * control_gate
                grad_output_gate = grad_h * np.tanh(all_cell_states[t, 0, idx, :])
                grad_forget_gate = grad_c * (
                    all_cell_states[t - 1, 0, idx, :]
                    if t > 0
                    else initial_cell_state[0, idx, :]
                    if initial_cell_state is not None
                    else 0
                )
                grad_control_gate = grad_c * input_gate

                grad_input_activation = grad_input_gate * input_gate * (1 - input_gate)
                grad_output_activation = grad_output_gate * output_gate * (1 - output_gate)
                grad_forget_activation = grad_forget_gate * forget_gate * (1 - forget_gate)
                grad_control_activation = grad_control_gate * (1 - control_gate**2)

                grad_inputs[t, idx, :] = (
                    np.dot(grad_input_activation, weights_i)
                    + np.dot(grad_output_activation, weights_o)
                    + np.dot(grad_forget_activation, weights_f)
                    + np.dot(grad_control_activation, weights_c)
                )
                grad_h = (
                    np.dot(grad_input_activation, rweights_i)
                    + np.dot(grad_output_activation, rweights_o)
                    + np.dot(grad_forget_activation, rweights_f)
                    + np.dot(grad_control_activation, rweights_c)
                )
                if t == 0 and grad_initial_hidden_state is not None:
                    grad_initial_hidden_state[0, idx, :] = grad_h

                grad_weights[0, : self._hidden_size, :] += np.dot(
                    np.expand_dims(grad_input_activation, axis=0).T, np.expand_dims(current_input, axis=0)
                )
                grad_weights[0, self._hidden_size : 2 * self._hidden_size, :] += np.dot(
                    np.expand_dims(grad_output_activation, axis=0).T, np.expand_dims(current_input, axis=0)
                )
                grad_weights[0, 2 * self._hidden_size : 3 * self._hidden_size, :] += np.dot(
                    np.expand_dims(grad_forget_activation, axis=0).T, np.expand_dims(current_input, axis=0)
                )
                grad_weights[0, 3 * self._hidden_size :, :] += np.dot(
                    np.expand_dims(grad_control_activation, axis=0).T, np.expand_dims(current_input, axis=0)
                )

                prev_h = (
                    all_hidden_states[t - 1, 0, idx, :]
                    if t > 0
                    else initial_hidden_state[0, idx, :]
                    if initial_hidden_state is not None
                    else 0
                )
                grad_recurrence_weights[0, : self._hidden_size, :] += np.dot(
                    np.expand_dims(grad_input_activation, axis=0).T, np.expand_dims(prev_h, axis=0)
                )
                grad_recurrence_weights[0, self._hidden_size : 2 * self._hidden_size, :] += np.dot(
                    np.expand_dims(grad_output_activation, axis=0).T, np.expand_dims(prev_h, axis=0)
                )
                grad_recurrence_weights[0, 2 * self._hidden_size : 3 * self._hidden_size, :] += np.dot(
                    np.expand_dims(grad_forget_activation, axis=0).T, np.expand_dims(prev_h, axis=0)
                )
                grad_recurrence_weights[0, 3 * self._hidden_size :, :] += np.dot(
                    np.expand_dims(grad_control_activation, axis=0).T, np.expand_dims(prev_h, axis=0)
                )

                if grad_bias is not None:
                    grad_bias[0, : self._hidden_size] += grad_input_activation
                    grad_bias[0, self._hidden_size : 2 * self._hidden_size] += grad_output_activation
                    grad_bias[0, 2 * self._hidden_size : 3 * self._hidden_size] += grad_forget_activation
                    grad_bias[0, 3 * self._hidden_size : 4 * self._hidden_size] += grad_control_activation
                    grad_bias[0, 4 * self._hidden_size : 5 * self._hidden_size] += grad_input_activation
                    grad_bias[0, 5 * self._hidden_size : 6 * self._hidden_size] += grad_output_activation
                    grad_bias[0, 6 * self._hidden_size : 7 * self._hidden_size] += grad_forget_activation
                    grad_bias[0, 7 * self._hidden_size :] += grad_control_activation

                if grad_peephole_weights is not None:
                    grad_peephole_weights[0, : self._hidden_size] += grad_input_activation * (
                        all_cell_states[t - 1, 0, idx, :]
                        if t > 0
                        else initial_cell_state[0, idx, :]
                        if initial_cell_state is not None
                        else 0
                    )
                    grad_peephole_weights[0, self._hidden_size : 2 * self._hidden_size] += (
                        grad_output_activation * all_cell_states[t, 0, idx, :]
                    )
                    grad_peephole_weights[0, 2 * self._hidden_size : 3 * self._hidden_size] += (
                        grad_forget_activation
                        * (
                            all_cell_states[t - 1, 0, idx, :]
                            if t > 0
                            else initial_cell_state[0, idx, :]
                            if initial_cell_state is not None
                            else 0
                        )
                    )

                grad_c = grad_prev_c

        return tuple(
            [
                out
                for out in [
                    grad_inputs,
                    grad_weights,
                    grad_recurrence_weights,
                    grad_bias,
                    grad_initial_hidden_state,
                    grad_initial_cell_state,
                    grad_peephole_weights,
                ]
                if out is not None
            ]
        )

    def backward_graph(
        self,
        bias=True,
        sequence_lengths=False,
        initial_hidden_state=True,
        initial_cell_state=True,
        peephole_weights=False,
        final_hidden_state=False,
        final_cell_state=False,
    ):
        """Generate the ONNX graph for the backward pass of the LSTM operator."""
        inputs = helper.make_tensor_value_info(
            "inputs", TensorProto.FLOAT, [self._sequence_length, self._batch_size, self.input_size]
        )
        weights = helper.make_tensor_value_info(
            "weights", TensorProto.FLOAT, [self._num_directions, 4 * self._hidden_size, self.input_size]
        )
        recurrence_weights = helper.make_tensor_value_info(
            "recurrence_weights", TensorProto.FLOAT, [self._num_directions, 4 * self._hidden_size, self._hidden_size]
        )
        sequence_lengths = (
            helper.make_tensor_value_info("sequence_lengths", TensorProto.INT64, [self._batch_size])
            if sequence_lengths
            else None
        )
        initial_hidden_state = (
            helper.make_tensor_value_info(
                "initial_hidden_state", TensorProto.FLOAT, [self._num_directions, self._batch_size, self._hidden_size]
            )
            if initial_hidden_state
            else None
        )
        initial_cell_state = (
            helper.make_tensor_value_info(
                "initial_cell_state", TensorProto.FLOAT, [self._num_directions, self._batch_size, self._hidden_size]
            )
            if initial_cell_state
            else None
        )

        all_hidden_states = helper.make_tensor_value_info(
            "all_hidden_states",
            TensorProto.FLOAT,
            [self._sequence_length, self._num_directions, self._batch_size, self._hidden_size],
        )
        final_hidden_state = (
            helper.make_tensor_value_info(
                "final_hidden_state", TensorProto.FLOAT, [self._num_directions, self._batch_size, self._hidden_size]
            )
            if final_hidden_state
            else None
        )
        final_cell_state = (
            helper.make_tensor_value_info(
                "final_cell_state", TensorProto.FLOAT, [self._num_directions, self._batch_size, self._hidden_size]
            )
            if final_cell_state
            else None
        )
        all_cell_states = helper.make_tensor_value_info(
            "all_cell_states",
            TensorProto.FLOAT,
            [self._sequence_length, self._num_directions, self._batch_size, self._hidden_size],
        )
        iofc = helper.make_tensor_value_info(
            "iofc",
            TensorProto.FLOAT,
            [self._sequence_length, self._num_directions, self._batch_size, 4 * self._hidden_size],
        )
        grad_all_hidden_states = helper.make_tensor_value_info(
            "grad_all_hidden_states",
            TensorProto.FLOAT,
            [self._sequence_length, self._num_directions, self._batch_size, self._hidden_size],
        )
        grad_final_hidden_state = (
            helper.make_tensor_value_info(
                "grad_final_hidden_state",
                TensorProto.FLOAT,
                [self._num_directions, self._batch_size, self._hidden_size],
            )
            if final_hidden_state
            else None
        )
        grad_final_cell_state = (
            helper.make_tensor_value_info(
                "grad_final_cell_state", TensorProto.FLOAT, [self._num_directions, self._batch_size, self._hidden_size]
            )
            if final_cell_state
            else None
        )

        grad_inputs = helper.make_tensor_value_info(
            "grad_inputs", TensorProto.FLOAT, [self._sequence_length, self._batch_size, self.input_size]
        )
        grad_weights = helper.make_tensor_value_info(
            "grad_weights", TensorProto.FLOAT, [self._num_directions, 4 * self._hidden_size, self.input_size]
        )
        grad_recurrence_weights = helper.make_tensor_value_info(
            "grad_recurrence_weights",
            TensorProto.FLOAT,
            [self._num_directions, 4 * self._hidden_size, self._hidden_size],
        )
        grad_bias = (
            helper.make_tensor_value_info("grad_bias", TensorProto.FLOAT, [self._num_directions, 8 * self._hidden_size])
            if bias
            else None
        )
        grad_initial_hidden_state = (
            helper.make_tensor_value_info(
                "grad_initial_hidden_state",
                TensorProto.FLOAT,
                [self._num_directions, self._batch_size, self._hidden_size],
            )
            if initial_hidden_state
            else None
        )
        grad_initial_cell_state = (
            helper.make_tensor_value_info(
                "grad_initial_cell_state",
                TensorProto.FLOAT,
                [self._num_directions, self._batch_size, self._hidden_size],
            )
            if initial_cell_state
            else None
        )
        grad_peephole_weights = (
            helper.make_tensor_value_info(
                "grad_peephole_weights", TensorProto.FLOAT, [self._num_directions, 3 * self._hidden_size]
            )
            if peephole_weights
            else None
        )

        lstm = helper.make_node(
            "LSTMGrad",
            inputs=[
                "inputs",
                "weights",
                "recurrence_weights",
                "sequence_lengths" if sequence_lengths is not None else "",
                "initial_hidden_state" if initial_hidden_state is not None else "",
                "initial_cell_state" if initial_cell_state is not None else "",
                "all_hidden_states" if all_hidden_states is not None else "",
                "all_cell_states" if all_cell_states is not None else "",
                "iofc" if iofc is not None else "",
                "grad_all_hidden_states" if grad_all_hidden_states is not None else "",
                "grad_final_hidden_state" if grad_final_hidden_state is not None else "",
                "grad_final_cell_state" if grad_final_cell_state else "",
            ],
            outputs=[
                "grad_inputs",
                "grad_weights",
                "grad_recurrence_weights",
                "grad_bias" if grad_bias is not None else "",
                "grad_initial_hidden_state" if grad_initial_hidden_state is not None else "",
                "grad_initial_cell_state" if grad_initial_cell_state is not None else "",
                "grad_peephole_weights" if grad_peephole_weights is not None else "",
            ],
            domain="com.microsoft",
            hidden_size=self._hidden_size,
        )

        graph = helper.make_graph(
            [lstm],
            "lstm",
            [
                gi
                for gi in [
                    inputs,
                    weights,
                    recurrence_weights,
                    sequence_lengths,
                    initial_hidden_state,
                    initial_cell_state,
                    all_hidden_states,
                    all_cell_states,
                    iofc,
                    grad_all_hidden_states,
                    grad_final_hidden_state,
                    grad_final_cell_state,
                ]
                if gi
            ],
            [
                go
                for go in [
                    grad_inputs,
                    grad_weights,
                    grad_recurrence_weights,
                    grad_bias,
                    grad_initial_hidden_state,
                    grad_initial_cell_state,
                    grad_peephole_weights,
                ]
                if go
            ],
        )

        self._backward_model = helper.make_model(
            graph,
            producer_name="orttraining",
            opset_imports=[helper.make_opsetid("", 14), helper.make_opsetid("com.microsoft", 1)],
        )

        return self._backward_model

    def backward_ort(
        self,
        inputs,
        weights,
        recurrence_weights,
        initial_hidden_state=None,
        initial_cell_state=None,
        all_hidden_states=None,
        all_cell_states=None,
        iofc=None,
        grad_all_hidden_states=None,
        grad_final_hidden_state=None,
        grad_final_cell_state=None,
    ):
        """Run LSTM backward using ONNX Runtime.

        Users must call backward_graph before calling this function.
        """
        ort_inputs = {
            "inputs": inputs,
            "weights": weights,
            "recurrence_weights": recurrence_weights,
            "all_hidden_states": all_hidden_states,
            "all_cell_states": all_cell_states,
            "iofc": iofc,
            "grad_all_hidden_states": grad_all_hidden_states,
        }
        if initial_hidden_state is not None:
            ort_inputs["initial_hidden_state"] = initial_hidden_state
        if initial_cell_state is not None:
            ort_inputs["initial_cell_state"] = initial_cell_state
        if grad_final_hidden_state is not None:
            ort_inputs["grad_final_hidden_state"] = grad_final_hidden_state
        if grad_final_cell_state is not None:
            ort_inputs["grad_final_cell_state"] = grad_final_cell_state

        ort_outs = None
        with tempfile.TemporaryDirectory() as tmpdirname:
            model_path = os.path.join(tmpdirname, "lstm_gradient.onnx")
            onnx.save(self._backward_model, model_path)
            ort_session = ort.InferenceSession(model_path, providers=["CPUExecutionProvider"])
            ort_outs = ort_session.run(None, ort_inputs)
        return ort_outs


@pytest.mark.parametrize("sequence_length", [2, 4, 16, 32])
@pytest.mark.parametrize("batch_size", [32])
@pytest.mark.parametrize("input_size", [32])
@pytest.mark.parametrize("hidden_size", [32])
def test_lstm_forward(sequence_length, batch_size, input_size, hidden_size):
    num_directions = 1

    lstm = LSTM(sequence_length, batch_size, input_size, hidden_size)
    _ = lstm.forward_graph(
        bias=True, sequence_lengths=False, initial_hidden_state=True, initial_cell_state=True, peephole_weights=False
    )

    inputs = np.random.rand(sequence_length, batch_size, input_size).astype(np.float32)
    weights = np.random.rand(num_directions, 4 * hidden_size, input_size).astype(np.float32)
    recurrence_weights = np.random.rand(num_directions, 4 * hidden_size, hidden_size).astype(np.float32)
    bias = np.random.rand(num_directions, 8 * hidden_size).astype(np.float32)
    initial_hidden_state = np.random.rand(num_directions, batch_size, hidden_size).astype(np.float32)
    initial_cell_state = np.random.rand(num_directions, batch_size, hidden_size).astype(np.float32)
    peephole_weights = None

    outs_ort = lstm.forward_ort(
        inputs, weights, recurrence_weights, bias, initial_hidden_state, initial_cell_state, peephole_weights
    )
    outs_np = lstm.forward_np(
        inputs, weights, recurrence_weights, bias, initial_hidden_state, initial_cell_state, peephole_weights
    )

    for ort_out, np_out in zip(outs_ort, outs_np, strict=False):
        assert np.allclose(ort_out, np_out, rtol=1e-03, atol=1e-05)


@pytest.mark.parametrize("sequence_length", [2, 4, 16, 32])
@pytest.mark.parametrize("batch_size", [32])
@pytest.mark.parametrize("input_size", [32])
@pytest.mark.parametrize("hidden_size", [32])
def test_lstm_backward(sequence_length, batch_size, input_size, hidden_size):
    num_directions = 1

    lstm = LSTM(sequence_length, batch_size, input_size, hidden_size)
    _ = lstm.backward_graph(
        bias=True,
        sequence_lengths=False,
        initial_hidden_state=True,
        initial_cell_state=True,
        peephole_weights=False,
        final_hidden_state=True,
        final_cell_state=True,
    )

    inputs = np.random.rand(sequence_length, batch_size, input_size).astype(np.float32)
    weights = np.random.rand(num_directions, 4 * hidden_size, input_size).astype(np.float32)
    recurrence_weights = np.random.rand(num_directions, 4 * hidden_size, hidden_size).astype(np.float32)
    bias = True
    initial_hidden_state = np.random.rand(num_directions, batch_size, hidden_size).astype(np.float32)
    initial_cell_state = np.random.rand(num_directions, batch_size, hidden_size).astype(np.float32)
    peephole_weights = False

    all_hidden_states = np.random.rand(sequence_length, num_directions, batch_size, hidden_size).astype(np.float32)
    all_cell_states = np.random.rand(sequence_length, num_directions, batch_size, hidden_size).astype(np.float32)
    iofc = np.random.rand(sequence_length, num_directions, batch_size, 4 * hidden_size).astype(np.float32)
    grad_all_hidden_states = np.random.rand(sequence_length, num_directions, batch_size, hidden_size).astype(np.float32)
    grad_final_hidden_state = np.random.rand(num_directions, batch_size, hidden_size).astype(np.float32)
    grad_final_cell_state = np.random.rand(num_directions, batch_size, hidden_size).astype(np.float32)

    outs_ort = lstm.backward_ort(
        inputs,
        weights,
        recurrence_weights,
        initial_hidden_state,
        initial_cell_state,
        all_hidden_states,
        all_cell_states,
        iofc,
        grad_all_hidden_states,
        grad_final_hidden_state,
        grad_final_cell_state,
    )
    outs_np = lstm.backward_np(
        inputs,
        weights,
        recurrence_weights,
        bias,
        initial_hidden_state,
        initial_cell_state,
        peephole_weights,
        all_hidden_states,
        all_cell_states,
        iofc,
        grad_all_hidden_states,
        grad_final_hidden_state,
        grad_final_cell_state,
    )

    for ort_out, np_out in zip(outs_ort, outs_np, strict=False):
        assert np.allclose(ort_out, np_out, rtol=1e-03, atol=1e-05)



import numpy as np
import onnx
from onnx import TensorProto, helper


def GenerateModel(model_name):  # noqa: N802
    nodes = [
        helper.make_node("Gather", ["embed_weights", "input_1"], ["gather_out"], "gather"),
        helper.make_node("Add", ["gather_out", "add_q_weight"], ["add_q_out"], "add_q"),
        helper.make_node("Add", ["gather_out", "add_k_weight"], ["add_k_out"], "add_k"),
        helper.make_node("Add", ["gather_out", "add_v_weight"], ["add_v_out"], "add_v"),
        helper.make_node(
            "Concat",
            ["add_q_out", "add_k_out", "add_v_out"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node("Add", ["add_qkv_weight", "concat_out"], ["add_out"], "add"),
        helper.make_node(
            "ReduceSum",
            ["add_out"],
            ["predictions"],
            "reduce_sum_1",
            axes=[0],
            keepdims=1,
        ),
    ]

    embed_weights = np.random.uniform(-1, 1, 8000).tolist()

    add_q_weight = [
        -0.23681640625,
        -0.16552734375,
        0.2191162109375,
        -0.1756591796875,
        -0.03460693359375,
        -0.05316162109375,
        -0.336181640625,
        -0.253662109375,
    ]

    add_k_weight = [
        0.0246734619140625,
        0.011993408203125,
        0.0178375244140625,
        0.00998687744140625,
        0.0255126953125,
        0.076416015625,
        -0.040771484375,
        0.0107879638671875,
    ]

    add_v_weight = [
        -0.005893707275390625,
        -0.00916290283203125,
        0.04541015625,
        0.0159454345703125,
        -0.0029163360595703125,
        -0.03472900390625,
        0.0535888671875,
        0.0091094970703125,
    ]

    initializers = [  # initializers
        helper.make_tensor("embed_weights", TensorProto.FLOAT, [1000, 8], embed_weights),
        helper.make_tensor("add_q_weight", TensorProto.FLOAT, [8], add_q_weight),
        helper.make_tensor("add_k_weight", TensorProto.FLOAT, [8], add_k_weight),
        helper.make_tensor("add_v_weight", TensorProto.FLOAT, [8], add_v_weight),
        helper.make_tensor("add_qkv_weight", TensorProto.FLOAT, [1], [1.0]),
    ]

    graph = helper.make_graph(
        nodes,
        "ConcatThreeInputs",  # name
        [helper.make_tensor_value_info("input_1", TensorProto.INT64, ["batch", "seq_len"])],  # inputs
        [  # outputs
            helper.make_tensor_value_info("predictions", TensorProto.FLOAT, [1, 1, 8]),
        ],
        initializers,
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


GenerateModel("concat_trainable.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------


class Row:
    """A row in a PTable"""

    def __init__(self, columns: list[str]) -> None:
        self._columns: list[str] = columns  # List of strings
        self._annotation_table = None  # Optional PTable used for displaying detailed information about the feature row.

    def append_annotation_table(self, ptable) -> None:
        self._annotation_table = ptable


class PTable:
    """A table that can be printed to the console."""

    def __init__(self, sortable=False) -> None:
        self._rows: list[Row] = []
        self._column_count = None
        self._sortable = sortable  # allow the rows to be sorted by the first column

    def add_row(self, columns: list[str]) -> Row:
        """Add a row to the table. The number of columns must match the number of columns in the table."""
        if self._column_count is None:
            self._column_count = len(columns)
        assert self._column_count == len(columns)
        row = Row(columns)
        self._rows.append(row)
        return row

    def get_string(self, first_column_width=None, second_column_width=None) -> str:
        """Serialize the table to a string."""
        if len(self._rows) == 0:
            return ""

        # Collect the max width of each column
        column_widths = []
        for row in self._rows:
            if column_widths:
                assert len(column_widths) == len(row._columns)
            else:
                column_widths = [0] * len(row._columns)
            for i, column in enumerate(row._columns):
                column_widths[i] = max(column_widths[i], len(str(column)))

        if first_column_width:
            column_widths[0] = max(first_column_width, column_widths[0])

        if second_column_width:
            column_widths[2] = max(second_column_width, column_widths[2])

        serialized_table = ""
        if self._sortable:
            sorted_rows = sorted(self._rows, key=lambda row: row._columns[0])
        else:
            sorted_rows = self._rows

        for row in sorted_rows:
            for i, column in enumerate(row._columns):
                serialized_table += f"{str(column).ljust(column_widths[i] + 2)}"
            serialized_table += "\n"
            if row._annotation_table:
                serialized_table += row._annotation_table.get_string(
                    first_column_width=column_widths[0], second_column_width=column_widths[2]
                )

        return serialized_table



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import copy
import warnings
from collections import OrderedDict, abc
from collections.abc import Mapping, Sequence

import torch

from onnxruntime.training.utils.torch_profile_utils import nvtx_function_decorator


class PrimitiveType:
    """Helper class for Python primitive types."""

    _primitive_types = {int, bool, float}  # noqa: RUF012

    @staticmethod
    def is_primitive_type(value):
        """Check if `value` is a Python primitive type."""
        return type(value) in PrimitiveType._primitive_types

    @staticmethod
    def get_tensor(value, device) -> torch.Tensor:
        """Convert `value` to a torch.Tensor."""
        return torch.tensor(value, device=device)

    @staticmethod
    def get_primitive_dtype(value):
        # If `value` is a boolean, save the value of the boolean in dtype.
        # This way, if the value changes from one forward call to the next, the schema will mismatch,
        # and the model will be re-exported.
        return f"{type(value)!s}_{value}" if isinstance(value, bool) else str(type(value))


# Data types supported as model inputs and outputs.
ORTModelInputOutputType = (
    str
    | int
    | bool
    | float
    | torch.Tensor
    | Sequence["ORTModelInputOutputType"]
    | Mapping[str, "ORTModelInputOutputType"]
    | None
)


class _TensorStub:
    """Tensor stub class used to represent model's input or output"""

    __slots__ = ["dtype", "name", "shape", "shape_dims", "tensor_idx"]

    def __init__(
        self,
        tensor_idx: int,
        name: str | None = None,
        dtype: str | None = None,
        shape=None,
        shape_dims: int | None = None,
    ):
        self.tensor_idx = tensor_idx
        self.name: str | None = name
        self.dtype: str | None = dtype
        self.shape = shape
        self.shape_dims: int | None = shape_dims  # r.g. rank.

    def __repr__(self) -> str:
        result = "_TensorStub("

        list_of_str = []
        if self.tensor_idx is not None:
            list_of_str.append(f"tensor_idx={self.tensor_idx}")
        if self.name is not None:
            list_of_str.append(f"name={self.name}")
        if self.dtype is not None:
            list_of_str.append(f"dtype={self.dtype}")
        if self.shape is not None:
            list_of_str.append(f"shape={self.shape}")
        if self.shape_dims is not None:
            list_of_str.append(f"shape_dims={self.shape_dims}")

        result += ",".join(list_of_str) + ")"
        return result

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, other):
        if not other:
            return False
        elif not isinstance(other, _TensorStub):
            raise NotImplementedError("_TensorStub must only be compared to another _TensorStub instance!")
        elif self.tensor_idx != other.tensor_idx:
            return False
        elif self.name != other.name:
            return False
        elif self.dtype != other.dtype:
            return False
        elif self.shape != other.shape:
            return False
        elif self.shape_dims != other.shape_dims:
            return False
        return True


# Data schema used to represent model's input or output.
ORTModelInputOutputSchemaType = (
    str | _TensorStub | Sequence["ORTModelInputOutputSchemaType"] | Mapping[str, "ORTModelInputOutputSchemaType"] | None
)


def _warn_of_constant_inputs(data):
    warnings.warn(
        f"Received input of type {type(data)} which may be treated as a constant by ORT by default."
        " Please consider moving constant arguments to the model constructor."
    )


@nvtx_function_decorator
def extract_data_and_schema(
    data: ORTModelInputOutputType, constant_as_tensor=False, device: torch.device | None = None
) -> tuple[list[torch.Tensor], ORTModelInputOutputSchemaType]:
    """Extract the data schema by replacing every torch.Tensor value with _TensorStub, and return all tensors in
    a list.

    Depth first traversal to iterate over the data:
    > Replace every tensor with a stub
    > Replace None/str typed data with itself
    > For other primitive types:
        If constant_as_tensor is True, create tensor from data , and replace them with a stub in returned schema;
        Otherwise, replace them with themselves in returned schema.

    Examples:
        Example 1, list:
            data = [torch.tensor(1), torch.tensor(2)]
            schema = [_TensorStub(shape=()), _TensorStub(shape=())]

        Example 2, dict:
            data = {"a": torch.tensor(1), "b": torch.tensor(2)}
            schema = {"a": _TensorStub(shape=()), "b": _TensorStub(shape=())}

        Example 3, dict of list:
            data = {"a": [torch.tensor(1), torch.tensor(2)], "b": [torch.tensor(3), torch.tensor(4)]}
            schema = {"a": [_TensorStub(shape=()), _TensorStub(shape=())],
                        "b": [_TensorStub(shape=()), _TensorStub(shape=())]}

        Example 4, nested dict:
            data = {"a": {"b": torch.tensor(1), "c": torch.tensor(2)},
                    "d": {"e": torch.tensor(3), "f": torch.tensor(4)}}
            schema = {"a": {"b": _TensorStub(shape=()), "c": _TensorStub(shape=())},
                        "d": {"e": _TensorStub(shape=()), "f": _TensorStub(shape=())}}

        Example 5, dict of mixed list and dict:
            data = {"a": [torch.tensor(1), torch.tensor(2)], "b": {"c": torch.tensor(3), "d": torch.tensor(4)}}
            schema = {"a": [_TensorStub(shape=()), _TensorStub(shape=())],
                        "b": {"c": _TensorStub(shape=()), "d": _TensorStub(shape=())}}

    Args:
        data: The data to extract the schema from, which can be in any kind of nested structure,
            including Sequence and Mapping.
        constant_as_tensor: Whether to treat constant inputs as tensor. Default is False.
        device: The device to create tensor for the constant.


    Returns:
        Tuple:
            The first value: a list of tensors extracted from the data.
            The second value: schema of the data, which has the same structure as the data.


    """

    flatten_tensor_data = []
    tensor_idx = [-1]

    def _flatten_from_data(data: ORTModelInputOutputType, prefix_name: str = ""):
        if data is None:
            return data
        elif isinstance(data, str):
            _warn_of_constant_inputs(data)
            return data
        elif PrimitiveType.is_primitive_type(data):
            _warn_of_constant_inputs(data)
            if constant_as_tensor:
                tensor_idx[0] += 1
                flatten_tensor_data.append(PrimitiveType.get_tensor(data, device))
                return _TensorStub(
                    tensor_idx[0], dtype=PrimitiveType.get_primitive_dtype(data), shape_dims=0, name=prefix_name
                )
            return data
        # Depth first traversal to iterate over the data to replace every tensor with a stub
        elif isinstance(data, torch.Tensor):
            tensor_idx[0] += 1
            flatten_tensor_data.append(data)
            return _TensorStub(
                tensor_idx[0],
                dtype=str(data.dtype),
                shape_dims=len(data.size()),
                name=prefix_name,
            )
        elif isinstance(data, abc.Sequence):
            sequence_type = type(data)
            stubbed_schema = []
            for i, val in enumerate(data):
                stubbed_schema.append(_flatten_from_data(val, f"{prefix_name}_{i}" if prefix_name else f"{i}"))

            try:
                # namedtuple can be created by passing the list sequence to method _make
                stubbed_schema = sequence_type._make(stubbed_schema)
            except AttributeError:
                # If attribute error is encountered, create the sequence directly
                stubbed_schema = sequence_type(stubbed_schema)
            return stubbed_schema
        elif isinstance(data, abc.Mapping):
            dict_type = type(data)
            stubbed_schema = OrderedDict()
            for key, val in data.items():
                stubbed_schema[key] = _flatten_from_data(val, f"{prefix_name}_{key}" if prefix_name else f"{key}")
            stubbed_schema = dict_type(**stubbed_schema)
            return stubbed_schema
        else:
            raise TypeError(f"Unsupported flatten data type: {type(data)}")

    schemas = _flatten_from_data(data)
    return flatten_tensor_data, schemas


@nvtx_function_decorator
def unflatten_data_using_schema(
    data: list[torch.Tensor], schema: ORTModelInputOutputSchemaType
) -> ORTModelInputOutputType:
    """Follows the schema to generate an output that is expected by the user.

    Depth first traversal to iterate over the schema:
    > Replace every _TensorStub with a tensor from data. For each _TensorStub, the tensor is selected from `data`
      by its tensor_idx.

    Examples:
        Example 1, schema is a list:
            data = [torch.tensor(1), torch.tensor(2)]
            schema = [_TensorStub(shape=()), True, _TensorStub(shape=()), 1, 2.0, "abc"]
            output = [torch.tensor(1), True, torch.tensor(2), 1, 2.0, "abc"]

        Example 2, schema is a dict:
            data = [torch.tensor(1), torch.tensor(2)]
            schema = {"a": _TensorStub(shape=()), "b": _TensorStub(shape=()), "c": True, "d": 1, "e": 2.0, "f": "abc"}
            output = {"a": torch.tensor(1), "b": torch.tensor(2), "c": True, "d": 1, "e": 2.0, "f": "abc"}

        Example 3, schema is a dict of list:
            data = [torch.tensor(1), torch.tensor(2), torch.tensor(3), torch.tensor(4)]
            schema = {"a": [_TensorStub(shape=()), _TensorStub(shape=())], "b": [_TensorStub(shape=()), _TensorStub(shape=())]}
            output = {"a": [torch.tensor(1), torch.tensor(2)], "b": [torch.tensor(3), torch.tensor(4)]}

        Example 4, schema is a nested dict:
            data = [torch.tensor(1), torch.tensor(2), torch.tensor(3), torch.tensor(4)]
            schema = {"a": {"b": _TensorStub(shape=()), "c": _TensorStub(shape=())},
                        "d": {"e": _TensorStub(shape=()), "f": _TensorStub(shape=())}}
            output = {"a": {"b": torch.tensor(1), "c": torch.tensor(2)}, "d": {"e": torch.tensor(3), "f": torch.tensor(4)}}

        Example 5, dict of mixed list and dict:
            data = [torch.tensor(1), torch.tensor(2), torch.tensor(3), torch.tensor(4)]
            schema = {"a": [_TensorStub(shape=()), _TensorStub(shape=())], "b": {"c": _TensorStub(shape=()), "d": _TensorStub(shape=())}}
            output = {"a": [torch.tensor(1), torch.tensor(2)], "b": {"c": torch.tensor(3), "d": torch.tensor(4)}}


    Args:
        data (List[torch.Tensor]): List of tensors to be used to replace the _TensorStub in schema
        schema (ORTModelInputOutputSchemaType): Schema to follow to generate the output

    Returns:
        output (ORTModelInputOutputType): Output that is expected by the user

    """

    def _replace_stub_with_tensor_value(data_schema: ORTModelInputOutputSchemaType, data: list[torch.Tensor]):
        # Recursively traverse across user_output and replace all _TensorStub
        # with torch.Tensor values from outputs following output_idx

        if data_schema is None:
            return None
        elif isinstance(data_schema, str):
            return data_schema
        elif PrimitiveType.is_primitive_type(data_schema):
            return data_schema
        elif isinstance(data_schema, _TensorStub):
            assert isinstance(data[data_schema.tensor_idx], torch.Tensor), (
                f"Expecting torch.Tensor, got {type(data[data_schema.tensor_idx])}"
            )
            return data[data_schema.tensor_idx]
        elif isinstance(data_schema, abc.Sequence):
            sequence_type = type(data_schema)
            if hasattr(sequence_type, "_make"):  # namedtuple
                sequence_type = type(data_schema)
                data_schema = sequence_type._make(_replace_stub_with_tensor_value(uo, data) for uo in data_schema)
            else:
                data_schema = sequence_type(_replace_stub_with_tensor_value(uo, data) for uo in data_schema)
            return data_schema
        elif isinstance(data_schema, abc.Mapping):
            new_user_output = copy.copy(data_schema)
            for key, schema_val in data_schema.items():
                new_user_output[key] = _replace_stub_with_tensor_value(schema_val, data)
            data_schema = new_user_output

            return data_schema

        raise TypeError(f"Unsupported unflatten data type: {type(data_schema)}.")

    user_output = _replace_stub_with_tensor_value(schema, data)
    return user_output



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

# -*- coding: UTF-8 -*-
# Taken from https://github.com/onnx/onnxmltools/blob/master/tests/end2end/test_custom_op.py.
import unittest

import numpy as np
import onnxmltools
from keras import Sequential
from keras import backend as K
from keras.layers import Conv2D, Layer, MaxPooling2D

import onnxruntime as onnxrt


class ScaledTanh(Layer):
    def __init__(self, alpha=1.0, beta=1.0, **kwargs):
        super().__init__(**kwargs)
        self.alpha = alpha
        self.beta = beta

    def build(self, input_shape):
        super().build(input_shape)

    def call(self, x):
        return self.alpha * K.tanh(self.beta * x)

    def compute_output_shape(self, input_shape):
        return input_shape


def custom_activation(scope, operator, container):
    # type:(ScopeBase, OperatorBase, ModelContainer) -> None
    container.add_node(
        "ScaledTanh",
        operator.input_full_names,
        operator.output_full_names,
        op_version=1,
        alpha=operator.original_operator.alpha,
        beta=operator.original_operator.beta,
    )


class TestInferenceSessionKeras(unittest.TestCase):
    def test_run_model_conv(self):
        # keras model
        N, C, H, W = 2, 3, 5, 5  # noqa: N806
        x = np.random.rand(N, H, W, C).astype(np.float32, copy=False)

        model = Sequential()
        model.add(
            Conv2D(
                2,
                kernel_size=(1, 2),
                strides=(1, 1),
                padding="valid",
                input_shape=(H, W, C),
                data_format="channels_last",
            )
        )
        model.add(ScaledTanh(0.9, 2.0))
        model.add(MaxPooling2D((2, 2), strides=(2, 2), data_format="channels_last"))

        model.compile(optimizer="sgd", loss="mse")
        actual = model.predict(x)
        self.assertIsNotNone(actual)

        # conversion
        converted_model = onnxmltools.convert_keras(model, custom_conversion_functions={ScaledTanh: custom_activation})
        self.assertIsNotNone(converted_model)

        # runtime
        content = converted_model.SerializeToString()
        rt = onnxrt.InferenceSession(content, providers=onnxrt.get_available_providers())
        input = {rt.get_inputs()[0].name: x}
        actual_rt = rt.run(None, input)
        self.assertEqual(len(actual_rt), 1)
        np.testing.assert_allclose(actual, actual_rt[0], rtol=1e-05, atol=1e-08)


if __name__ == "__main__":
    unittest.main(module=__name__, buffer=True)
